/*
 * This file is part of SessionJ.
 *
 * Copyright (C) Raymond Hu, Imperial College London.
 *
 */

package sj;

import java.util.*;
import java.io.*;

import polyglot.ast.*;
import polyglot.frontend.*;
import polyglot.frontend.goals.*;
import polyglot.lex.Lexer;
import polyglot.main.*;
import polyglot.types.*;
import polyglot.util.*;
import polyglot.visit.*;

import sj.ast.*;
import sj.parse.Lexer_c;
import sj.parse.Grm;
import sj.types.*;
import sj.visit.*;

/**
 * Extension information for sj extension.
 */
public class ExtensionInfo extends polyglot.frontend.JLExtensionInfo
{
	static
	{
		// force Topics to load
		Topics t = new Topics();
	}

	public String defaultFileExtension()
	{
		return "sj";
	}

	public String compilerName()
	{
		return "sjc";
	}

	public Parser parser(Reader reader, FileSource source, ErrorQueue eq)
	{
		Lexer lexer = new Lexer_c(reader, source, eq);
		Grm grm = new Grm(lexer, ts, nf, eq);
		return new CupParser(grm, source, eq);
	}

	protected NodeFactory createNodeFactory()
	{
		return new SJNodeFactory_c();
	}

	protected TypeSystem createTypeSystem()
	{
		return new SJTypeSystem_c();
	}

	// Must override the constructor to get this visitor to run!
	public Scheduler createScheduler()
	{
		return new SJScheduler(this);
	}

	// Check polyglot.frontend.Scheduler for the ordered list of abstract goals (but still unsure where the barriers are)
	static class SJScheduler extends JLScheduler
	{
		SJScheduler(ExtensionInfo extInfo)
		{
			super(extInfo);
		}

		// TypesInitialized phase.
		public Goal SJProcedureDeclParsing(final Job job)
		{
			TypeSystem ts = job.extensionInfo().typeSystem();
			NodeFactory nf = job.extensionInfo().nodeFactory();
			Goal g = internGoal(new VisitorGoal(job, new SJProcedureDeclParser(job, ts, nf))
			{
				public Collection prerequisiteGoals(Scheduler scheduler)
				{
					List l = new ArrayList();
					l.add(scheduler.Parsed(job)); // Copied from inside polyglot.frontend.goals.TypesInitialized (i.e. the next pass).
					return l;
				}
			});
			return g;
		}

    public Goal TypesInitialized(Job job)
    {
			TypeSystem ts = extInfo.typeSystem();
			NodeFactory nf = extInfo.nodeFactory();
			Goal g = internGoal(new VisitorGoal(job, new TypeBuilder(job, ts, nf))
			{
				public Collection prerequisiteGoals(Scheduler scheduler)
				{
					List l = new ArrayList();
					l.addAll(super.prerequisiteGoals(scheduler));
					l.add(SJProcedureDeclParsing(job));
					return l;
	    	}
			});
			return g;
    }

    /*public Goal TypeChecked(Job job)
    {
			TypeSystem ts = extInfo.typeSystem();
			NodeFactory nf = extInfo.nodeFactory();
			Goal g = internGoal(new VisitorGoal(job, new TypeChecker(job, ts, nf))
			{
				public Collection prerequisiteGoals(Scheduler scheduler)
				{
					List l = new ArrayList();
					l.add(scheduler.Disambiguated(job));
					l.addAll(super.prerequisiteGoals(scheduler));
					return l;
				}

				public Collection corequisiteGoals(Scheduler scheduler)
				{
					List l = new ArrayList();
					l.add(scheduler.ConstantsChecked(job));
					l.addAll(super.corequisiteGoals(scheduler));
					return l;
				}
			});
			return g;
    }*/

		// ReachabilityChecked phase (post TypeChecked).
		public Goal SJCreateOperationParsing(final Job job)
		{
			TypeSystem ts = job.extensionInfo().typeSystem();
			NodeFactory nf = job.extensionInfo().nodeFactory();
			Goal g = internGoal(new VisitorGoal(job, new SJCreateOperationParser(job, ts, nf))
			{
				public Collection prerequisiteGoals(Scheduler scheduler)
				{
					List l = new ArrayList();
					l.add(scheduler.TypeChecked(job)); // Copied from ReachabilityChecked pass.
					l.add(scheduler.ConstantsChecked(job));
					return l;
				}
			});
			return g;
		}

		public Goal SJSessionOperationParsing(final Job job)
		{
			TypeSystem ts = job.extensionInfo().typeSystem();
			NodeFactory nf = job.extensionInfo().nodeFactory();
			Goal g = internGoal(new VisitorGoal(job, new SJSessionOperationParser(job, ts, nf))
			{
				public Collection prerequisiteGoals(Scheduler scheduler)
				{
					List l = new ArrayList();
					//l.addAll(super.prerequisiteGoals(scheduler));
					l.add(SJCreateOperationParsing(job));
					return l;
				}
			});
			return g;
		}

		public Goal SJSessionTypeBuilding(final Job job)
		{
			TypeSystem ts = job.extensionInfo().typeSystem();
			NodeFactory nf = job.extensionInfo().nodeFactory();
			Goal g = internGoal(new VisitorGoal(job, new SJSessionTypeBuilder(job, ts, nf))
			{
				public Collection prerequisiteGoals(Scheduler scheduler)
				{
					List l = new ArrayList();
					//l.addAll(super.prerequisiteGoals(scheduler));
					l.add(SJSessionOperationParsing(job));
					return l;
				}
			});
			return g;
		}

		/*public Goal SJSessionTypeBuilding2(final Job job)
		{
			TypeSystem ts = job.extensionInfo().typeSystem();
			NodeFactory nf = job.extensionInfo().nodeFactory();
			Goal g = internGoal(new VisitorGoal(job, new SJSessionTypeBuilder2(job, ts, nf))
			{
				public Collection prerequisiteGoals(Scheduler scheduler)
				{
					List l = new ArrayList();
					//l.addAll(super.prerequisiteGoals(scheduler));
					l.add(SJSessionTypeBuilding(job));
					return l;
				}
			});
			return g;
		}*/

		public Goal SJSessionTypeChecking(final Job job)
		{
			TypeSystem ts = job.extensionInfo().typeSystem();
			NodeFactory nf = job.extensionInfo().nodeFactory();
			Goal g = internGoal(new VisitorGoal(job, new SJSessionTypeChecker(job, ts, nf))
			{
				public Collection prerequisiteGoals(Scheduler scheduler)
				{
					List l = new ArrayList();
					//l.addAll(super.prerequisiteGoals(scheduler));
					l.add(SJSessionTypeBuilding(job));
					//l.add(SJSessionTypeBuilding2(job));
					return l;
				}
			});
			return g;
		}

		public Goal SJAliasChecking(final Job job) // Maybe should be before type checking, but session-receive would not yet be typed (does it matter?).
		{
			TypeSystem ts = job.extensionInfo().typeSystem();
			NodeFactory nf = job.extensionInfo().nodeFactory();
			Goal g = internGoal(new VisitorGoal(job, new SJAliasChecker(job, ts, nf))
			{
				public Collection prerequisiteGoals(Scheduler scheduler)
				{
					List l = new ArrayList();
					//l.addAll(super.prerequisiteGoals(scheduler));

					if (SJConstants.SJ_TYPES_CHECKED)
					{
						l.add(SJSessionTypeChecking(job));
					}
					else
					{
						l.add(SJSessionTypeBuilding(job));
					}

					return l;
				}
			});
			return g;
		}

		public Goal ReachabilityChecked(final Job job)
		{
			TypeSystem ts = extInfo.typeSystem();
			NodeFactory nf = extInfo.nodeFactory();
			Goal g = internGoal(new ReachabilityChecked(job, ts, nf)
			{
				public Collection prerequisiteGoals(Scheduler scheduler)
				{
					List l = new ArrayList();
					l.addAll(super.prerequisiteGoals(scheduler));
					l.add(SJAliasChecking(job));
					return l;
				}
			});
			return g;
    }

		// Serialized phase.
		public Goal SJProtocolDeclTranslation(final Job job)
		{
			TypeSystem ts = extInfo.typeSystem();
			NodeFactory nf = extInfo.nodeFactory();
			Goal g = internGoal(new VisitorGoal(job, new SJProtocolDeclTranslator(job, ts, nf))
			{
				public Collection prerequisiteGoals(Scheduler scheduler)
				{
					List l = new ArrayList();
					l.add(scheduler.ReachabilityChecked(job)); // Copied from Serialized pass.
					l.add(scheduler.ExceptionsChecked(job));
					l.add(scheduler.ExitPathsChecked(job));
					l.add(scheduler.InitializationsChecked(job));
					l.add(scheduler.ConstructorCallsChecked(job));
					l.add(scheduler.ForwardReferencesChecked(job));
					//l.addAll(super.prerequisiteGoals(scheduler));
					return l;
				}
			});
			return g;
		}

		public Goal SJStructuralOperationTranslation(final Job job)
		{
			TypeSystem ts = extInfo.typeSystem();
			NodeFactory nf = extInfo.nodeFactory();
			Goal g = internGoal(new VisitorGoal(job, new SJStructuralOperationTranslator(job, ts, nf))
			{
				public Collection prerequisiteGoals(Scheduler scheduler)
				{
					List l = new ArrayList();
					l.add(SJProtocolDeclTranslation(job));
					return l;
				}
			});
			return g;
		}

		public Goal SJProcedureDeclTranslation(final Job job)
		{
			TypeSystem ts = extInfo.typeSystem();
			NodeFactory nf = extInfo.nodeFactory();
			Goal g = internGoal(new VisitorGoal(job, new SJProcedureDeclTranslator(job, ts, nf))
			{
				public Collection prerequisiteGoals(Scheduler scheduler)
				{
					List l = new ArrayList();
					l.add(SJStructuralOperationTranslation(job));
					return l;
				}
			});
			return g;
		}

		public Goal SJThreadTranslation(final Job job)
		{
			TypeSystem ts = job.extensionInfo().typeSystem();
			NodeFactory nf = job.extensionInfo().nodeFactory();
			Goal g = internGoal(new VisitorGoal(job, new SJThreadTranslator(job, ts, nf))
			{
				public Collection prerequisiteGoals(Scheduler scheduler)
				{
					List l = new ArrayList();
					l.add(SJProcedureDeclTranslation(job));
					return l;
				}
			});
			return g;
		}

		public Goal SJSessionTryTranslation(final Job job)
		{
			TypeSystem ts = extInfo.typeSystem();
			NodeFactory nf = extInfo.nodeFactory();
			Goal g = internGoal(new VisitorGoal(job, new SJSessionTryTranslator(job, ts, nf))
			{
				public Collection prerequisiteGoals(Scheduler scheduler)
				{
					List l = new ArrayList();
					l.add(SJThreadTranslation(job));
					return l;
				}
			});
			return g;
		}

		public Goal SJHigherOrderTranslation(final Job job)
		{
			TypeSystem ts = extInfo.typeSystem();
			NodeFactory nf = extInfo.nodeFactory();
			Goal g = internGoal(new VisitorGoal(job, new SJHigherOrderTranslator(job, ts, nf))
			{
				public Collection prerequisiteGoals(Scheduler scheduler)
				{
					List l = new ArrayList();
					l.add(SJSessionTryTranslation(job));
					return l;
				}
			});
			return g;
		}

		public Goal Serialized(final Job job)
		{
			Goal g = internGoal(new Serialized(job)
			{
				public Collection prerequisiteGoals(Scheduler scheduler)
				{
					List l = new ArrayList();
					l.addAll(super.prerequisiteGoals(scheduler));
					l.add(SJHigherOrderTranslation(job));
					return l;
				}
			});
			return g;
		}
	}
}
