/*
 * This file is part of SessionJ.
 *
 * Copyright (C) Raymond Hu, Imperial College London.
 *
 */

package sj.ast.sessops.structuralops;

import polyglot.ast.*;
import polyglot.types.*;
import polyglot.visit.*;
import polyglot.util.*;

import java.util.*;

import sj.extension.SJStructuralOperationExt;
import sj.util.SJLabel;

import static sj.SJConstants.SJ_KEYWORD_INBRANCH;

/**
 * An "expression-less switch"
 */
public class SJInbranch_c extends Stmt_c implements SJInbranch
{
	protected List<SJInbranchCase> branchCases;
	private List<Expr> arguments;

	public SJInbranch_c(Position pos, List<SJInbranchCase> branchCases, List<Expr> arguments)
	{
		super(pos);

		this.branchCases = TypedList.copyAndCheck(branchCases, SJInbranchCase.class, true); // Is this check necessary?
		this.arguments = arguments;
	}

	public List<SJInbranchCase> branchCases()
	{
		return Collections.unmodifiableList(this.branchCases); // Why unmodifiable?
	}

	public SJInbranch branchCases(List<SJInbranchCase> branchCases)
	{
		SJInbranch_c n = (SJInbranch_c) copy();

		n.branchCases = TypedList.copyAndCheck(branchCases, SJInbranchCase.class, true); // Generic mixed up with Polyglot.

		return n;
	}

	protected SJInbranch_c reconstruct(List<SJInbranchCase> branchCases)
	{
		if (!CollectionUtil.equals(branchCases, this.branchCases))
		{
			SJInbranch_c n = (SJInbranch_c) copy();

			n.branchCases = TypedList.copyAndCheck(branchCases, SJInbranchCase.class, true);

			return n;
		}

		return this;
	}

	public Context enterScope(Context c)
	{
		return c.pushBlock();
	}

	public Node visitChildren(NodeVisitor v)
	{
		List branchCases = visitList(this.branchCases, v);
		return reconstruct(branchCases);
	}

	// Children checked by typechecker before visiting this node (from leaveCall).
	public Node typeCheck(TypeChecker tc) throws SemanticException
	{
		Collection<SJLabel> labels = new HashSet<SJLabel>();

		for (SJInbranchCase ibc : branchCases)
		{
			SJLabel lab = ibc.label();

			if (labels.contains(lab))
			{
				throw new SemanticException("Duplicate case label: " + lab + ".", ibc.position());
			}

			labels.add(lab);
		}
		return this;
	}

	public String toString()
	{
		return ((SJStructuralOperationExt) ext()).target() + "." + SJ_KEYWORD_INBRANCH + "() { ... }";
	}

	public void prettyPrint(CodeWriter w, PrettyPrinter tr) // This is largely redundant (except for debugging) due to SJInbranchTranslator
	{
		w.write(((SJStructuralOperationExt) ext()).target() + "." + SJ_KEYWORD_INBRANCH + "()"); // Should use the SJCompilerUtils operation.
		w.write(" {");
		w.allowBreak(4, " ");
		w.begin(0);

		for (Iterator<SJInbranchCase> i = branchCases.iterator(); i.hasNext(); )
		{
			SJInbranchCase s = (SJInbranchCase) i.next();
			w.allowBreak(4, " ");
			print(s, w, tr);
		}

		w.end();
		w.allowBreak(0, " ");
		w.write("}");
	}

	public Term entry()
	{
		return ((NamedVariable) ((SJStructuralOperationExt) ext()).target()).entry();
	}

	public List acceptCFG(CFGBuilder v, List succs)
	{
		List<Term> caseEntries = new LinkedList<Term>();

		for (SJInbranchCase ibc : branchCases)
		{
			caseEntries.add(ibc.entry());
		}

		v.visitCFG(((NamedVariable) ((SJStructuralOperationExt) ext()).target()), FlowGraph.EDGE_KEY_OTHER, caseEntries); // entry...
		v.visitCFGList(branchCases, this); // ...and exit points?

		return succs;
	}

	public List<Expr> arguments()
	{
		return arguments;
	}
}
