/*
 * This file is part of SessionJ.
 *
 * Copyright (C) Raymond Hu, Imperial College London.
 *
 */

package sj.util;

import java.util.*;

import polyglot.ast.*;
import polyglot.types.*;

import sj.ast.*;
import sj.ast.sessvars.SJSocketVariable;
import sj.ast.protocoldecls.SJProtocolDecl;
import sj.ast.typenodes.SJTypeNode;
import sj.ast.sessvars.*;
import sj.ast.sessops.*;
import sj.ast.sessops.basicops.*;
import sj.ast.sessops.structuralops.SJStructuralOperation;
import sj.ast.servops.*;
import sj.ast.sesscasts.*;
import sj.ast.proceduredecls.*;
import sj.ast.sessformals.*;
import sj.extension.*;
import sj.extension.protocoldecls.*;
import sj.extension.sessops.*;
import sj.extension.sesscasts.*;
import sj.extension.sessformals.*;
import sj.types.typeobjects.sesstypes.SJType;
import sj.types.typeobjects.typeableinstances.*;

import static sj.SJConstants.*;

public final class SJCompilerUtils
{
	private SJCompilerUtils() { }

	public static void debugPrintln(Object obj)
	{
		if (SJ_DEBUG_MODE)
		{
			System.out.println(obj);
		}
	}

	public static List<SJType> getSJFormalTypes(SJMethodInstance sjmi)
	{
		List<SJType> sjFormalTypes = new ArrayList<SJType>();

		SJMap sjFormals = sjmi.sjFormals();

		for (String sjFormal : sjFormals.keySet())
		{
			sjFormalTypes.add(sjFormals.get(sjFormal)); // SJMethodInstance preserves arg. order.
		}

		return sjFormalTypes;
	}

	public static String parseSJReturnName(Node parent) throws SemanticException
	{
		String returned;

		if (parent instanceof LocalDecl)
		{
			returned = ((LocalDecl) parent).name();
		}
		else if (parent instanceof Assign)
		{
			returned = ((NamedVariable) ((Assign) parent).left()).name();
		}
		else
		{
			throw new SemanticException("Invalid assignment of session-return operation (anonymous return not done yet): " + parent);
		}

		return returned;
	}

	/**
	* Static setters and getters for AST information - want to avoid writing out the extension object cast every time. Can't put into AST nodes because of inheritance restrictions (using Polyglot base classes) and would defeat the point to put into the extension object itself.
	*/

	// AST (directly on the Ext object).
	public static SJTypeNode getSJProtocolDeclTypeNode(SJProtocolDecl n)
	{
		return ((SJProtocolDeclExt) n.ext()).sjTypeNode();
	}

	public static SJProtocolDecl setSJProtocolDeclTypeNode(SJProtocolDecl n, SJTypeNode sjTypeNode)
	{
		return ((SJProtocolDeclExt) n.ext()).sjTypeNode(sjTypeNode);
	}

	public static String getSJNamedName(SJNamed n)
	{
		return ((SJNamedExt) n.ext()).sjTypeableName();
	}

	public static SJNamed setSJNamedName(SJNamed n, String name)
	{
		return ((SJNamedExt) n.ext()).sjTypeableName(name);
	}

	public static Receiver getSJStructuralOpTarget(SJStructuralOperation sjso)
	{
		return ((SJStructuralOperationExt) sjso.ext()).target();
	}

	public static SJStructuralOperation setSJStructuralOpTarget(SJStructuralOperation sjso, SJSocketVariable sv)
	{
		return ((SJStructuralOperationExt) sjso.ext()).target(sv);
	}

	public static SJTypeNode getSJTypeCastTypeNode(SJTypeCast n)
	{
		return ((SJTypeCastExt) n.ext()).sjTypeNode();
	}

	public static String getSJReturnName(SJReturnOperation sro)
	{
		SJReturnOperationExt roe;

		if (sro instanceof SJBasicOperation) // Channel/session-receive.
		{
			roe = (SJReturnOperationExt) ((SJBasicOperation) sro).ext().ext();
		}
		else //if (sro instanceof SJAccept)
		{
			roe = (SJReturnOperationExt) ((SJAccept) sro).ext().ext();
		}

		return roe.returned();
	}

	public static SJReturnOperation setSJReturnName(SJReturnOperation sro, String returned)
	{
		SJReturnOperationExt roe;

		if (sro instanceof SJBasicOperation)
		{
			roe = (SJReturnOperationExt) ((SJBasicOperation) sro).ext().ext();
		}
		else //if (sro instanceof SJAccept)
		{
			roe = (SJReturnOperationExt) ((SJAccept) sro).ext().ext();
		}

		return roe.returned(returned);
	}

	public static SJTypeNode getSJTypeFormalTypeNode(SJTypeFormal n)
	{
		return ((SJTypeFormalExt) n.ext()).sjTypeNode();
	}

	public static SJTypeFormal setSJTypeFormalTypeNode(SJTypeFormal n, SJTypeNode sjTypeNode)
	{
		return ((SJTypeFormalExt) n.ext()).sjTypeNode(sjTypeNode);
	}

	public static List<SJFormal> getSJMethodDeclFormals(SJMethodDecl n)
	{
		return ((SJMethodDeclExt) n.ext()).sjFormals();
	}

	public static SJTypeNode getSJMethodDeclReturnType(SJMethodDecl n)
	{
		return ((SJMethodDeclExt) n.ext()).sjReturnType();
	}

	public static SJMethodDecl setSJMethodDeclReturnType(SJMethodDecl n, SJTypeNode returnType)
	{
		return ((SJMethodDeclExt) n.ext()).sjReturnType(returnType);
	}

	public static String getSJCallReturnName(SJCall n)
	{
		return ((SJCallExt) n.ext()).returned();
	}

	public static SJCall setSJCallReturnName(SJCall n, String returned)
	{
		return ((SJCallExt) n.ext()).returned(returned);
	}

	public static List<SJLocalSocket> getSJCallArgs(SJCall n)
	{
		return ((SJCallExt) n.ext()).socketArgs();
	}

	// Typeable instances. // Try to combine/reduce these?
	public static SJNamedInstance getSJNamedInstance(SJNamed n)
	{
		return ((SJNamedExt) n.ext()).sjNamedInstance();
	}

	public static SJNamed setSJNamedInstance(SJNamed n, SJNamedInstance sjni)
	{
		return ((SJNamedExt) n.ext()).sjNamedInstance(sjni);
	}

	public static SJSessionTryInstance getSJSessionTryInstance(SJSessionTry n)
	{
		return ((SJSessionTryExt) n.ext()).sjSessionTryInstance();
	}

	public static SJSessionTry setSJSessionTryInstance(SJSessionTry n, SJSessionTryInstance sjsti)
	{
		return ((SJSessionTryExt) n.ext()).sjSessionTryInstance(sjsti);
	}

	public static SJSessionOpInstance getSJSessionOpInstance(SJSessionOperation n)
	{
		return ((SJSessionOperationExt) n.ext()).sjSessionOpInstance();
	}

	public static SJSessionOperation setSJSessionOpInstance(SJSessionOperation n, SJSessionOpInstance sjsoi)
	{
		return ((SJSessionOperationExt) n.ext()).sjSessionOpInstance(sjsoi);
	}

	public static SJNamedInstance getSJReturnOpInstance(SJReturnOperation n)
	{
		if (n instanceof SJBasicOperation) // Channel/session-receive.
		{
			return ((SJReturnOperationExt) ((SJBasicOperation) n).ext().ext()).sjNamedInstance();
		}
		else //if (n instanceof SJAccept)
		{
			return ((SJReturnOperationExt) ((SJAccept) n).ext().ext()).sjNamedInstance();
		}
	}

	public static SJReturnOperation setSJReturnOpInstance(SJReturnOperation n, SJNamedInstance sjni)
	{
		if (n instanceof SJBasicOperation)
		{
			return ((SJReturnOperationExt) ((SJBasicOperation) n).ext().ext()).sjNamedInstance(sjni);
		}
		else //if (n instanceof SJAccept)
		{
			return ((SJReturnOperationExt) ((SJAccept) n).ext().ext()).sjNamedInstance(sjni);
		}
	}

	public static SJMethodInstance getSJMethodInstance(SJMethodDecl n)
	{
		return ((SJMethodDeclExt) n.ext()).sjMethodInstance();
	}

	public static SJMethodDecl setSJMethodInstance(SJMethodDecl n, SJMethodInstance sjmi)
	{
		return ((SJMethodDeclExt) n.ext()).sjMethodInstance(sjmi);
	}

	public static SJMethodInstance getSJCallMethodInstance(SJCall n)
	{
		return ((SJCallExt) n.ext()).sjMethodInstance();
	}

	public static SJCall setSJCallMethodInstance(SJCall n, SJMethodInstance sjmi)
	{
		return ((SJCallExt) n.ext()).sjMethodInstance(sjmi);
	}

	public static SJThreadInstance getSJSpawnThreadInstance(SJSpawn n)
	{
		return ((SJSpawnExt) n.ext()).sjThreadInstance();
	}

	public static SJSpawn setSJSpawnThreadInstance(SJSpawn n, SJThreadInstance sjti)
	{
		return ((SJSpawnExt) n.ext()).sjThreadInstance(sjti);
	}

	// Utility (values on the typeable instances).
	public static String getSessionName(SJSessionOperation n)
	{
		return ((SJSessionOperationExt) n.ext()).session(); // From the AST: may be better to use type information. But maybe need to use before types have been built?
	}

	public static String getServerName(SJServerOperation n)
	{
		return ((SJServerOperationExt) n.ext()).server();
	}

	public static SJType getSJType(SJTypeable n)
	{
		return ((SJTypeableExt) n.ext()).sjTypeableInstance().sjType();
	}

	public static SJSessionOperation setSJType(SJSessionOperation n, SJType sjType)
	{
		SJSessionOperationExt soe = (SJSessionOperationExt) n.ext();

		return soe.sjSessionOpInstance(soe.sjSessionOpInstance().sjType(sjType));
	}

	public static SJNamed setSJNamedType(SJNamed n, SJType sjType)
	{
		SJNamedExt sjne = (SJNamedExt) n.ext();

		return sjne.sjNamedInstance(sjne.sjNamedInstance().sjType(sjType));
	}

	public static boolean anonymousSJReturn(SJReturnOperation sjro)
	{
		SJReturnOperationExt roe;

		if (sjro instanceof SJBasicOperation) // Channel/session-receive.
		{
			roe = (SJReturnOperationExt) ((SJBasicOperation) sjro).ext().ext();
		}
		else //if (sjro instanceof SJAccept)
		{
			roe = (SJReturnOperationExt) ((SJAccept) sjro).ext().ext();
		}

		return roe.anonymousReturn();
	}

	public static boolean anonymousSJReturn(SJCall sjc)
	{
		return ((SJCallExt) sjc.ext()).anonymousReturn();
	}

	public static SJType getSJCallReturnType(SJCall n)
	{
		return ((SJCallExt) n.ext()).sjMethodInstance().sjReturnType();
	}

	public static List<Receiver> getSJSpawnArgs(SJSpawn sjs)
	{
		return ((SJSpawnExt) sjs.ext()).socketArgs();
	}

	public static SJSpawn setSJSpawnArgs(SJSpawn sjs, List<Receiver> socketArgs)
	{
		return ((SJSpawnExt) sjs.ext()).socketArgs(socketArgs);
	}

	public static Type subsumeSendTypes(Type type1, Type type2) throws SemanticException
	{
		if (type1 instanceof SJType)
		{
			if (!(type2 instanceof SJType))
			{
				throw new SemanticException("Not subsumable: " + type1 + ", " + type2);
			}

			return ((SJType) type1).subsume((SJType) type2);
		}
		else if (type1 instanceof ReferenceType)
		{
			if (!(type2 instanceof ReferenceType))
			{
				throw new SemanticException("Not subsumable: " + type1 + ", " + type2);
			}

			if (type1.isSubtype(type2)) // Could be checked earlier.
			{
				return type2;
			}
			else if (type2.isSubtype(type1))
			{
				return type1;
			}
			else
			{
				throw new SemanticException("Full send type subsumption not done yet.");
			}
		}
		else if (type1.typeEquals(type2))
		{
			return type1;
		}

		throw new SemanticException("Not subsumable: " + type1 + ", " + type2);
	}

	public static Type subsumeReceiveTypes(Type type1, Type type2) throws SemanticException
	{
		if (type1 instanceof SJType)
		{
			if (!(type2 instanceof SJType))
			{
				throw new SemanticException("Not subsumable: " + type1 + ", " + type2);
			}

			return ((SJType) type1).subsume((SJType) type2);
		}
		else if (type1 instanceof ReferenceType)
		{
			if (!(type2 instanceof ReferenceType))
			{
				throw new SemanticException("Not subsumable: " + type1 + ", " + type2);
			}

			if (type1.isSubtype(type2))
			{
				return type1;
			}
			else if (type2.isSubtype(type1))
			{
				return type2;
			}
			else
			{
				throw new SemanticException("Full receive type subsumption not done yet.");
			}
		}
		else if (type1.typeEquals(type2))
		{
			return type1;
		}

		throw new SemanticException("Not subsumable: " + type1 + ", " + type2);
	}

	/*public static int getSVUID(Type messageType) throws SemanticException // Simpler version of getMessageSVUID in SJRuntimeUtils.
	{
		int svuid;

		try // Doesn't work if the class binary is not available a priori.
		{
			String typeName = fullTypeName(messageType);

			if (messageType instanceof SJType)
			{
				if (messageType instanceof SJUnknownType)
				{
					// Can't do anything with the current (lack of) information.
				}
				else
				{
					// Useful to assign some special value?
				}
			}
			else if (messageType.isClass())
			{
				Class c = Class.forName(typeName);

				if (!c.isInterface()) // Interfaces don't have svuids.
				{
					ObjectStreamClass osc = ObjectStreamClass.lookup(c);

					if (osc == null)
					{
						throw new SemanticException("Class not serializable: " + c);
					}

					svuid = osc.getSerialVersionUID();
				}
			}
			else
			{
				// Primitives, arrays...

				throw new RuntimeException("Non-class types not yet done: " + messageType);
			}
		}
		catch (ClassNotFoundException cnfe)
		{
			throw new SemanticException(cnfe);
		}

		return svuid;
	}

	public static String fullTypeName(Type type)
	{
		if (type instanceof SJType)
		{
			return type.toString();
		}
		else if (type.isPrimitive())
		{
			return type.toString();
		}
		else if (type.isClass())
		{
			return type.toClass().fullName();
		}
		else if (type.isArray())
		{
			throw new RuntimeException("Array types not done yet: " + type);
		}

		throw new RuntimeException("Shouldn't get here: " + type);
	}*/
}
