/*
 * This file is part of SessionJ.
 *
 * Copyright (C) Raymond Hu, Imperial College London.
 *
 */

package sj.visit;

import java.util.*;

import polyglot.ast.*;
import polyglot.frontend.Job;
import polyglot.types.*;
import polyglot.util.*;
import polyglot.visit.ErrorHandlingVisitor;
import polyglot.visit.NodeVisitor;

import sj.SJConstants;
import sj.ast.*;
import sj.ast.sessvars.SJLocalSocket;
import sj.ast.sessops.SJSpawn;
import sj.ast.sessops.basicops.*;
import sj.ast.proceduredecls.*;
import sj.ast.socketops.*;
import sj.ast.sessformals.*;
import sj.types.SJTypeSystem;
import sj.types.typeobjects.sesstypes.*;
import sj.util.*;

// Does type encoding like SJProtocolDeclTranslator, but simpler overall. Also does channel passing operations, SJCall, SJSpawn and SJReturn. // Works in conjunction with SJSessionTryTranslator (latter must go first).
public class SJHigherOrderTranslator extends ErrorHandlingVisitor
{
	private SJTypeSystem sjts = (SJTypeSystem) typeSystem();
	private SJNodeFactory sjnf = (SJNodeFactory) nodeFactory();

	private SJTypeEncoder sjte = new SJTypeEncoder(sjts);

	private Stack<Term> contexts = new Stack<Term>();

	public SJHigherOrderTranslator(Job job, TypeSystem ts, NodeFactory nf)
	{
		super(job, ts, nf);
	}

	protected NodeVisitor enterCall(Node parent, Node n) throws SemanticException
	{
		if (n instanceof SJMethodDecl)
		{
			contexts.push((SJMethodDecl) n);
		}
		else if (n instanceof SJSessionTry)
		{
			contexts.push((SJSessionTry) n);
		}

		return this;
	}

	protected Node leaveCall(Node old, Node n, NodeVisitor v) throws SemanticException
	{
		if (n instanceof SJMethodDecl || n instanceof SJSessionTry)
		{
			contexts.pop();
		}
		else if (n instanceof SJSendSession || n instanceof SJReceiveSession || n instanceof SJReceiveChannel) // Channel-send takes a channel argument which the protocol can be obtained from.
		{
			SJBasicOperation sjbo = (SJBasicOperation) n;
			SJType sjType = SJCompilerUtils.getSJType(sjbo); // Could instead get the return type directly from the SJReturnOperationExt.

			if (sjbo instanceof SJSendSession)
			{
				sjType = (SJType) ((SJSendType) sjType).messageType();
			}
			else //if (sjbo instanceof SJReceiveSession || sjbo instanceof SJReceiveChannel)
			{
				sjType = (SJType) ((SJReceiveType) sjType).messageType();
			}

			n = translateHOOperation(sjbo, sjType);
		}
		else if (n instanceof Eval)
		{
			Eval eval = (Eval) n;
			Expr expr = eval.expr();

			if (expr instanceof SJSpawn)
			{
				n = translateSJSpawn(eval, (SJSpawn) expr);
			}
			else if (expr instanceof SJCall)
			{
				n = translateSJCall(eval, (SJCall) expr);
			}
			else if (expr instanceof LocalDecl)
			{
				LocalDecl ld = (LocalDecl) expr;
				Expr init = ld.init();

				if (init instanceof SJCall)
				{
					throw new SemanticException("Local declarations initialised by session-method calls not done yet: " + ld);

					//n = translateSJCall(eval, (SJCall) init); // Current translation breaks the original scoping for the declared variable. // Break LocalDecls into uninit. decl. and assign?
				}
			}
			else if (expr instanceof Assign)
			{
				Assign ass = (Assign) expr;
				Expr right = ass.right();

				if (right instanceof SJCall)
				{
					n = translateSJCall(eval, (SJCall) right);
				}
			}
		}
		else if (n instanceof SJReturn)
		{
			n = translateSJReturn((SJReturn) n);
		}

		return n;
	}

	private SJBasicOperation translateHOOperation(SJBasicOperation sjbo, SJType sjType) throws SemanticException
	{
		Position pos = sjbo.position(); // Positions are messed up.

		List<Expr> callArgs = new LinkedList<Expr>();

		callArgs.add(sjnf.StringLit(pos, sjte.encode(sjType))); // Maybe can just pass encoded type to session-receive directly.

		//Call call = sjnf.Call(pos, sjbo.target(), sjnf.Id(pos, SJConstants.SJSOCKET_DECODESJTYPE), callArgs);

		List<Expr> sjboArgs = new LinkedList<Expr>();

		if (sjbo instanceof SJSendSession)
		{
			sjboArgs.add((Expr) sjbo.arguments().get(0)); // The session to be delegated. // Preserve original arguments to allow e.g. timeout parameters.
		}

		//sjboArgs.add(call);
		sjboArgs.addAll(callArgs); // Pass the encoded type directly, rather than decode it automatically.

		return (SJBasicOperation) sjbo.arguments(sjboArgs);
	}

	private Stmt translateSJSpawn(Eval eval, SJSpawn sjs)
	{
		List<Stmt> translated = new LinkedList<Stmt>();

		Id incrementSpawnCount = sjnf.Id(eval.position(), SJConstants.SJSOCKET_INCREMENTSPAWNCOUNT);

		translated.add(eval);

		for (Object obj : SJCompilerUtils.getSJSpawnArgs(sjs))
		{
			SJLocalSocket sv = (SJLocalSocket) obj;
			Position pos = sv.position();

			translated.add(sjnf.Eval(pos, sjnf.Call(pos, sv, incrementSpawnCount)));
			translated.add(sjnf.Eval(pos, sjnf.Assign(pos, sv, Assign.ASSIGN, sjnf.NullLit(pos)))); // Could also add this for session send and receive.
		}

		return sjnf.Block(sjs.position(), translated);
	}

	private Stmt translateSJCall(Eval eval, SJCall sjc)
	{
		List<Stmt> translated = new LinkedList<Stmt>();

		Id getSpawnCount = sjnf.Id(eval.position(), SJConstants.SJSOCKET_GETSPAWNCOUNT);

		int v = 0;

		for (SJLocalSocket sv : SJCompilerUtils.getSJCallArgs(sjc))
		{
			Position pos = sv.position();

			if (checkSJTryContext(sv))
			{
				Local local = sjnf.Local(pos, sjnf.Id(pos, SJConstants.SJ_SPAWN_COUNT_PREFIX + sv.name()));
				Call call = sjnf.Call(pos, sv, getSpawnCount);

				translated.add(v++, sjnf.Eval(pos, sjnf.Assign(pos, local, Assign.ASSIGN, call)));
				//translated.add(sjnf.Eval(pos, sjnf.Assign(pos, sv, Assign.ASSIGN, sjnf.NullLit(pos)))); // Incorrect.
			}
		}

		if (translated.size() == 0)
		{
			if (contexts.peek()	instanceof SJRecursiveMethodDecl) // Recursive session method calls inside a sessrec method can't be in a local session-try scope (so the above translation won't have been applied). // This translation routine was started in SJProcedureDeclTranslator. Maybe should move both parts to a separate pass.
			{
				SJRecursiveMethodDecl sjrmd = (SJRecursiveMethodDecl) contexts.peek();

				List<String> nestedRecursions = new LinkedList<String>(); // HACK: because the only visitor that knows if a session method call is a recursive one or not (i.e. has recurse operation types for the session arguments) is SJTypeChecker. So instead, we'll manually check through the method formal types to find nested recursions, and calls to one of those methods from this context are not recursive calls.

				for (SJFormal formal : SJCompilerUtils.getSJMethodDeclFormals(sjrmd)) // Enough to check for just one (recursive) argument?
				{
					SJType sjtype = SJCompilerUtils.getSJType(formal);

					if (sjtype instanceof SJRecursionType)
					{
						nestedRecursions = getNestedRecursions(((SJRecursionType) sjtype).body());

						break;
					}
				}

				Position pos = sjc.position();

				for (SJLocalSocket sv : SJCompilerUtils.getSJCallArgs(sjc))
				{
					if (!nestedRecursions.contains(sjc.name()))
					{
						SJSocketOperation sockop = sjnf.SJRecursionTrue(pos, sv, new SJLabel(sjc.name()));

						translated.add(sjnf.Eval(pos, sockop));
					}
				}

				translated.add(eval);

				return sjnf.Block(pos, translated);
			}

			return eval;
		}
		else
		{
			translated.add(v, eval);

			return sjnf.Block(sjc.position(), translated);
		}
	}

	private boolean checkSJTryContext(SJLocalSocket sv)
	{
		for (int i = contexts.size() - 1; i >= 0; i--)
		{
			Term term = contexts.get(i);

			if (term instanceof SJSessionTry)
			{
				for (Receiver rec : ((SJSessionTry) term).sockets())
				{
					if (((SJLocalSocket) rec).name().equals(sv.name()))
					{
						return true;
					}
				}
			}
			else if (term instanceof SJProcedureDecl)
			{
				return false;
			}
		}

		return false;
	}

	private Node translateSJReturn(SJReturn sjr)
	{
		if (contexts.peek() instanceof SJSessionTry) // Importantly, skips recursive session methods.
		{
			Expr expr = sjr.expr();
			Position pos = expr.position();

			Local returned = sjnf.Local(pos, sjnf.Id(pos, ((SJSessionTry) contexts.peek()).getReturned()));
			Assign ass = sjnf.Assign(pos, returned, Assign.ASSIGN, expr);

			return sjnf.Block(pos, sjnf.Eval(pos, ass), sjnf.Return(pos, returned));
		}
		else
		{
			return sjr;
		}
	}

	private List<String> getNestedRecursions(SJType sjtype) // HACK.
	{
		List<String> nestedRecursions = new LinkedList<String>();

		if (sjtype == null)
		{
			return nestedRecursions;
		}

		while (sjtype.child() != null)
		{
			sjtype = sjtype.child();
		}

		if (sjtype instanceof SJOutbranchType)
		{
			SJOutbranchType obt = (SJOutbranchType) sjtype;

			for (SJLabel lab : obt.labelSet())
			{
				nestedRecursions.addAll(getNestedRecursions(obt.branchCase(lab)));
			}
		}
		else if (sjtype instanceof SJInbranchType)
		{
			SJInbranchType ibt = (SJInbranchType) sjtype;

			for (SJLabel lab : ibt.labelSet())
			{
				nestedRecursions.addAll(getNestedRecursions(ibt.branchCase(lab)));
			}
		}
		else if (sjtype instanceof SJRecursionType)
		{
			SJRecursionType sjrt = (SJRecursionType) sjtype;

			nestedRecursions.add(sjrt.label().labelValue());
			nestedRecursions.addAll(getNestedRecursions(sjrt.body()));
		}

		return nestedRecursions;
	}
}
