/*
 * This file is part of SessionJ.
 *
 * Copyright (C) Raymond Hu, Imperial College London.
 *
 */

package sj.visit;

import java.util.*;

import polyglot.ast.*;
import polyglot.frontend.Job;
import polyglot.types.*;
import polyglot.util.*;
import polyglot.visit.*;

import sj.SJConstants;
import sj.ast.*;
import sj.ast.sessvars.*;
import sj.types.*;
import sj.util.*;

// Works in conjunction with SJHigherOrderTranslator.
public class SJSessionTryTranslator extends ErrorHandlingVisitor
{
	private SJTypeSystem sjts = (SJTypeSystem) typeSystem();
	private SJNodeFactory sjnf = (SJNodeFactory) nodeFactory();

	private SJCounter sjc = new SJCounter();

	private Type intType = sjts.Int();

	public SJSessionTryTranslator(Job job, TypeSystem ts, NodeFactory nf)
	{
		super(job, ts, nf);
	}

	protected Node leaveCall(Node old, Node n, NodeVisitor v) throws SemanticException
	{
		if (n instanceof SJSessionTry)
		{
			n = translateSJSessionTry((SJSessionTry) n);
		}

		return n;
	}

	private Block translateSJSessionTry(SJSessionTry sjst) throws SemanticException
	{
		Position pos = sjst.position(); // Position values are quite messed up.
		TypeNode intTypeNode = sjnf.CanonicalTypeNode(pos, intType);

		Id close = sjnf.Id(pos, SJConstants.SJSOCKET_CLOSE);

		List<Stmt> translated = new LinkedList<Stmt>();

		Id returned = sjnf.Id(pos, SJConstants.SJ_SESSION_TRY_RETURNED + sjc.nextValue());
		TypeNode socketType = sjnf.CanonicalTypeNode(pos, SJConstants.SJSOCKET_TYPE);

		translated.add(0, sjnf.LocalDecl(pos, Flags.NONE, socketType, returned, sjnf.NullLit(pos)));

		List<If> ifs = new LinkedList<If>();

		for (Receiver sock : sjst.sockets())
		{
			SJSocketVariable sjsv = (SJSocketVariable) sock;
			Id fname = sjnf.Id(pos, SJConstants.SJ_SPAWN_COUNT_PREFIX + sjsv.name());

			LocalDecl ld = sjnf.LocalDecl(pos, Flags.NONE, intTypeNode, fname, sjnf.IntLit(pos, IntLit.INT, SJConstants.SJ_SPAWN_COUNT_CLOSE));

			translated.add(ld);

			List<Expr> args = new LinkedList<Expr>();

			args.add(sjnf.Local(pos, fname));

			Expr cond1 = sjnf.Binary(pos, sjsv, Binary.NE, sjnf.NullLit(pos));
			Expr cond2 = sjnf.Binary(pos, sjsv, Binary.NE, sjnf.SJLocalSocket(pos, returned));
			Expr cond = sjnf.Binary(pos, cond1, Binary.COND_AND, cond2);

			Stmt consequent = sjnf.Eval(pos, sjnf.Call(pos, sjsv, close, args)); // Assuming close operation does not throw exceptions.

			ifs.add(sjnf.If(pos, cond, consequent));
		}

		Block finallyBlock = sjst.finallyBlock();

		if (finallyBlock == null)
		{
			sjst = (SJSessionTry) sjst.finallyBlock(sjnf.Block(pos, ifs));
		}
		else
		{
			for (If ifn : ifs)
			{
				finallyBlock = finallyBlock.append(ifn);
			}

			sjst = (SJSessionTry) sjst.finallyBlock(finallyBlock);
		}

		translated.add(sjst.setReturned(returned.id()));

		return sjnf.Block(pos, translated);
	}
}
