/*
 * This file is part of SessionJ.
 *
 * Copyright (C) Raymond Hu, Imperial College London.
 *
 */

package sj.visit;

import java.util.*;

import polyglot.ast.*;
import polyglot.frontend.Job;
import polyglot.types.*;
import polyglot.util.*;
import polyglot.visit.*;

import sj.SJConstants;
import sj.ast.*;
import sj.ast.sessvars.*;
import sj.ast.sessops.*;
import sj.ast.sessops.basicops.*;
import sj.ast.sessops.structuralops.*;
import sj.ast.servops.*;
import sj.ast.proceduredecls.*;
//import sj.ast.sessformals.*;
import sj.types.*;
import sj.types.typeobjects.sesstypes.*;
import sj.types.typeobjects.typeableinstances.*;
import sj.util.*;
import sj.visit.contexts.*;
import sj.visit.contexts.sesscontexts.*;

/**
 * Also does receive-cast insertion. And builds types for structural operations and finishes type building for higher-order operations (which includes session method-calls and session-spawns).
 */
public class SJSessionTypeChecker extends SJContextVisitor
{
	private SJTypeSystem sjts = (SJTypeSystem) typeSystem();
	private SJNodeFactory sjnf = (SJNodeFactory) nodeFactory();

	private SJTypeBuilder sjtb = new SJTypeBuilder(job(), sjts, sjnf);

	public SJSessionTypeChecker(Job job, TypeSystem ts, NodeFactory nf)
	{
		super(job, ts, nf);
	}

	protected SJContextVisitor sjEnterCall(Node parent, Node n) throws SemanticException
	{
		if (n instanceof SJSessionTry) // Could instead override the scope enter method (would probably be cleaner to do so, like in SJTypeBuilder). // But alright to do here because it is paired up with a corresponding routine in sjLeaveCall (which could similarly be done in the leave scope method instead).
		{
			SJSessionTryInstance sjsti = SJCompilerUtils.getSJSessionTryInstance((SJSessionTry) n);

			sjContext().pushSJSessionTry(sjsti.sockets());
		}
		else if (n instanceof SJStructuralOperation)
		{
			enterStructuralOperation((SJStructuralOperation) n);
		}
		else if (n instanceof SJInbranchCase)
		{
			sjContext().pushSJInbranchCase(((SJInbranchCase) n).label());
		}

		return this;
	}

	private void enterStructuralOperation(SJStructuralOperation sjso) throws SemanticException
	{
		String session = SJCompilerUtils.getSessionName(sjso);

		if (sjso instanceof SJOutbranch)
		{
			sjContext().pushSJOutbranch(session, ((SJOutbranch) sjso).label());
		}
		else if (sjso instanceof SJInbranch)
		{
			sjContext().pushSJInbranch(session);
		}
		else if (sjso instanceof SJOutwhile)
		{
			sjContext().pushSJOutwhile(session);
		}
		else if (sjso instanceof SJInwhile)
		{
			sjContext().pushSJInwhile(session);
		}
		else if (sjso instanceof SJRecursion)
		{
			sjContext().pushSJRecursion(session, ((SJRecursion) sjso).label());
		}
	}

	protected Node sjLeaveCall(Node parent, Node old, Node n, NodeVisitor v) throws SemanticException
	{
		if (n instanceof SJProcedureDecl)
		{
			checkSessionCompletion(); // No need to update outer context (will not be a session context), unlike SJSessionTry, which is why we can leave the scope management to SJContextVisitor parent class.
		}
		else if (n instanceof SJSessionTry) // Could be moved to an overriding leave scope method (see corresponding sjEnterCall comment).
		{
			checkSessionCompletion();

			sjContext().updateContext((SJSessionTryContext) sjContext().pop());
		}
		else if (n instanceof SJBasicOperation)
		{
			n = checkBasicOperation(parent, (SJBasicOperation) n);
		}
		else if (n instanceof SJStructuralOperation) // Should check there are no breaks or continues in while (and general iteration and recursion) scopes (unless all sessions have been necessarily completed).
		{
			n = checkStructuralOperation((SJStructuralOperation) n);
		}
		else if (n instanceof SJInbranchCase)
		{
			n = checkInbranchCase((SJInbranchCase) n);
		}
		else if (n instanceof SJServerOperation)
		{
			n = checkServerOperation((SJServerOperation) n);
		}
		else if (n instanceof SJCall)
		{
			n = checkSJCall(parent, (SJCall) n);
		}
		else if (n instanceof SJReturn)
		{
			n = checkSJReturn((SJReturn) n);
		}
		else if (n instanceof SJSpawn)
		{
			n = checkSJSpawn((SJSpawn) n);
		}

		return n;
	}

	private void checkSessionCompletion() throws SemanticException
	{
		for (String session : sjContext().sessionsInScope())
		{
			SJType expected = sjContext().expectedType(session); // Can just use activeType.

			if (expected != null)
			{
				throw new SemanticException("Session implementation incomplete for `" + session + "': implemented `" + sjContext().implementedType(session) + "', but required `" + expected + "'");
			}
		}
	}

	private Node checkBasicOperation(Node parent, SJBasicOperation sjbo) throws SemanticException
	{
		Node n = sjbo; // Node due to potential cast insertion for receive operations.

		String session = SJCompilerUtils.getSessionName(sjbo);
		SJType sjboType = SJCompilerUtils.getSJType(sjbo);

		if (sjbo instanceof SJRequest)
		{
			SJType sjType = sjContext().socketType(session).treeClone();

			sjContext().openSession(session, sjType); // Multiple sequential implementations of a session within same scope not allowed.

			SJCompilerUtils.debugPrintln("Opened: `" + session + "' " + sjType);
		}

		SJType expected = sjContext().expectedType(session);

		if (expected == null)
		{
			throw new SemanticException("Unexpected operation: " + sjbo);
		}
		else
		{
			expected = expected.nodeClone(); // To get just the current operation.
		}

		boolean typeError = false;

		if (sjbo instanceof SJRequest)
		{
			if (!(expected instanceof SJBeginType)) // Already checked at channel creation by type builder.
			{
				typeError = true;
			}
		}
		else if (sjbo instanceof SJSend || sjbo instanceof SJSendBoolean || sjbo instanceof SJSendInt || sjbo instanceof SJSendByte || sjbo instanceof SJSendChannel) // Make a send base class?
		{
			if (!(expected.isSubtype(sjboType))) // "Backwards" ("cancels out" SJType implementation).
			{
				typeError = true;
			}
		}
		else if (sjbo instanceof SJSendSession)
		{
			String arg = ((SJSocketVariable) sjbo.arguments().get(0)).name();

			if (arg.equals(session)) // This would never be type safe anyway.
			{
				throw new SemanticException("Cannot delegate a session over itself: " + sjbo);
			}

			SJType messageType = sjContext().activeType(arg);

			if (messageType == null)
			{
				sjboType = null; // To make error message more helpful.
				typeError = true; // Can't delegate inactive or completed session.
			}
			else
			{
				sjboType = sjts.SJSendType(messageType.treeClone()); // Should be ok for "further higher-order" types?

				if (!(sjboType.isSubtype(expected)))
				{
					typeError = true;
				}
				else
				{
					sjContext().delegateSession(arg);

					sjbo = (SJSendSession) SJCompilerUtils.setSJType(sjbo, sjboType);

					SJCompilerUtils.debugPrintln("Delegated: `" + arg + "' " + messageType);

					n = sjbo;
				}
			}
		}
		else if (sjbo instanceof SJReceive)
		{
			SJReceive sjr = (SJReceive) sjbo;

			if (((SJReceiveType) sjboType).messageType() instanceof SJUnknownType)
			{
				if (!(expected instanceof SJReceiveType)) // This could be pushed out earlier, but done here to keep the following case uniform with other operations.
				{
					typeError = true;
				}
				else
				{
					sjboType = expected.nodeClone();

					Type messageType = ((SJReceiveType) sjboType).messageType();
					Position pos = sjbo.position();

					if (messageType instanceof SJType)
					{
						String returned = SJCompilerUtils.parseSJReturnName(parent); // Anonymous session return caught here.
						SJSocketVariable target = sjr.target();
						SJType sjType = (SJType) messageType;

						SJReturnOperation sjro = null;

						if (messageType instanceof SJBeginType)
						{
							sjro = sjnf.SJReceiveChannel(pos, target, sjnf.Id(pos, SJConstants.SJSOCKET_RECEIVECHANNEL), (List<Expr>) sjr.arguments());

							sjContext().addChannel(sjts.SJNamedInstance(pos, sjType, returned));
						}
						else
						{
							sjro = sjnf.SJReceiveSession(pos, target, sjnf.Id(pos, SJConstants.SJSOCKET_RECEIVESESSION), (List<Expr>) sjr.arguments());

							sjContext().openSession(returned, sjType);
						}

						SJCompilerUtils.debugPrintln("Parsed: " + sjro);

						sjro = (SJReturnOperation) SJCompilerUtils.setSJSessionOpInstance((SJBasicOperation) sjro, sjts.SJSessionOpInstance(pos, sjboType, session)); // Operation type.
						sjro = (SJReturnOperation) SJCompilerUtils.setSJReturnOpInstance(sjro, sjts.SJNamedInstance(pos, sjType, returned)); // Return type.

						n = sjro;
					}
					else // Cast insertion. Relying on javac to re-type check the parent expression/statement (e.g. LocalDecl, LocalAssign, etc.) since the inserted cast may be incorrect. (Would be better to check ourselves - could run TypeChecker on new source.) // Inference for primitive-receives not possible: return type overloading not allowed (for receive).
					{
						sjr = (SJReceive) SJCompilerUtils.setSJType(sjr, sjboType); // Operation type.

						n = sjnf.Cast(pos, sjnf.CanonicalTypeNode(pos, messageType), sjr); // Also need to support array types.
					}

					SJCompilerUtils.debugPrintln("Built: `" + session + "' " + sjboType);

					n = sjtb.buildTypes(context(), n); // Base type information.
				}
			}
			else
			{
				if (!(sjboType.isSubtype(expected))) // "Backwards".
				{
					typeError = true;
				}
			}
		}
		else if (sjbo instanceof SJReceiveBoolean || sjbo instanceof SJReceiveInt || sjbo instanceof SJReceiveByte)
		{
			if (!(sjboType.typeEquals(expected)))
			{
				typeError = true;
			}
		}
		else if (sjbo instanceof SJReceiveChannel || sjbo instanceof SJReceiveSession)
		{
			SJReturnOperation sjro = (SJReturnOperation) sjbo;
			Position pos = sjbo.position();

			if (SJCompilerUtils.anonymousSJReturn(sjro)) // Would we get in here?
			{
				throw new SemanticException("Anonymous return operations not done yet: " + sjro);
			}

			String returned = SJCompilerUtils.getSJReturnName(sjro);

			if (((SJReceiveType) sjboType).messageType() instanceof SJUnknownType)
			{
				if (!(expected instanceof SJReceiveType) || !(((SJReceiveType) expected).messageType() instanceof SJType))
				{
					typeError = true;
				}
				else
				{
					sjboType = expected.nodeClone();

					SJType sjType = (SJType) ((SJReceiveType) sjboType).messageType();

					if (sjro instanceof SJReceiveChannel)
					{
						if (!sjType.wellFormed())
						{
							throw new SemanticException("Received channel type not well formed: " + sjType);
						}

						sjro = (SJReceiveChannel) SJCompilerUtils.setSJType((SJReceiveChannel) sjro, sjboType);

						sjContext().addChannel(sjts.SJNamedInstance(pos, sjType, returned));
					}
					else if (sjro instanceof SJReceiveSession)
					{
						if (!sjType.treeWellFormed()) // Includes check for bad 'begin' elements.
						{
							throw new SemanticException("Received session type not well formed: " + sjType);
						}

						sjro = (SJReceiveSession) SJCompilerUtils.setSJType((SJReceiveSession) sjro, sjboType);

						sjContext().openSession(returned, sjType);

						SJCompilerUtils.debugPrintln("Opened: `" + returned + "' " + sjType);
					}

					sjro = (SJReturnOperation) SJCompilerUtils.setSJReturnOpInstance(sjro, sjts.SJNamedInstance(sjro.position(), sjType, returned));

					SJCompilerUtils.debugPrintln("Built: `" + sjro + "' " + sjboType);

					n = sjro;
				}
			}
			else
			{
				if (!(sjboType.isSubtype(expected)))
				{
					typeError = true;
				}
				else
				{
					SJType sjType = (SJType) ((SJReceiveType) sjboType).messageType();

					if (sjro instanceof SJReceiveChannel)
					{
						if (!sjType.wellFormed())
						{
							throw new SemanticException("Received channel type not well formed: " + sjType);
						}

						sjContext().addChannel(sjts.SJNamedInstance(pos, sjType, returned));
					}
					else if (sjro instanceof SJReceiveSession)
					{
						if (!sjType.treeWellFormed())
						{
							throw new SemanticException("Received session type not well formed: " + sjType);
						}

						sjContext().openSession(returned, sjType);
					}

					n = sjro;
				}
			}
		}
		else if (sjbo instanceof SJRecurse)
		{
			if (!(sjboType.typeEquals(expected)))
			{
				typeError = true;
			}
		}

		if (typeError) // Replace this test with an exception factory call in each case? (Exception messages are less useful in the current way.)
		{
			throw new SemanticException("For `" + session + "' expected `" + expected.nodeToString() + "', not: " + sjboType);
		}

		sjContext().advanceSession(session, sjboType.nodeClone()); // treeClone should be ok.

		SJCompilerUtils.debugPrintln("Checked: `" + n + "' " + sjboType);

		return n;
	}

	private SJStructuralOperation checkStructuralOperation(SJStructuralOperation sjso) throws SemanticException
	{
		checkSessionCompletion(); // For inbranch, the target session has been manually advanced if all branch cases have been implemented.

		SJSessionContext sjsc = (SJSessionContext) sjContext().pop();
		String session = SJCompilerUtils.getSessionName(sjso);
		SJType implemented = sjsc.implementedType(session);
		SJType existing = SJCompilerUtils.getSJType(sjso); // Correct kind checked on scope enterance (type skeleton built by SJTypeBuilder).

		if (sjso instanceof SJOutbranch)
		{
			SJIOBranchCaseContext bcc = (SJIOBranchCaseContext) sjsc;

			implemented = ((SJOutbranchType) existing).branchCase(bcc.label(), implemented);

			bcc.removeSession(session); // Could be factored out into context object?

			sjContext().updateContext(bcc);
		}
		else if (sjso instanceof SJInbranch)
		{
			SJInbranchContext ibc = (SJInbranchContext) sjsc; // All the cases have already been built for the inbranch type by the inbranch-case checking routine.

			ibc.removeSession(session);

			sjContext().updateContext(ibc);
		}
		else if (sjso instanceof SJOutwhile) // While-contexts don't need to do updateContext (they shield outer session from nested operations).
		{
			implemented = ((SJOutwhileType) existing).body(implemented);
		}
		else if (sjso instanceof SJInwhile)
		{
			implemented = ((SJInwhileType) existing).body(implemented);
		}
		else if (sjso instanceof SJRecursion) // No updateContext (like in/outwhile).
		{
			implemented = ((SJRecursionType) existing).body(implemented);
		}

		sjso = (SJStructuralOperation) SJCompilerUtils.setSJSessionOpInstance(sjso, sjts.SJSessionOpInstance(sjso.position(), implemented, session));

		sjContext().advanceSession(session, implemented.nodeClone());

		SJCompilerUtils.debugPrintln("Checked: `" + session + "': " + implemented);

		return sjso;
	}

	private SJInbranchCase checkInbranchCase(SJInbranchCase ibc) throws SemanticException
	{
		checkSessionCompletion();

		SJIOBranchCaseContext bcc = (SJIOBranchCaseContext) sjContext().popBranchCaseContext();
		String session = bcc.session();

		SJType sjType = bcc.implementedType(session);
		SJInbranchType implemented = (SJInbranchType) sjContext().implementedType(session);

		if (implemented == null)
		{
			implemented = sjts.SJInbranchType().branchCase(bcc.label(), sjType);
		}
		else
		{
			implemented = implemented.branchCase(bcc.label(), sjType); // The opposite to our notion of "subsumption".
		}

		sjContext().implementedType(session, implemented);

		SJType active = sjContext().activeType(session);

		if (active != null && active.isSubtype(implemented)) // To satisfy session operation checking (session completion for structural operations) for inbranch (basically finishes advancing the inbranch session in the parent inbranch context when it is safe to do so, i.e. all cases have been implemented).
		{
			sjContext().activeType(session, null); // Also prevents re-implementing cases. But also prevents the implementation of extra cases, which is type-safe.
		}

		bcc.removeSession(session);

		return ibc;
	}

	private SJServerOperation checkServerOperation(SJServerOperation sjso) throws SemanticException
	{
		//String server = SJCompilerUtils.getServerName(sjso);
		SJType sjsoType = SJCompilerUtils.getSJType(sjso);

		if (sjso instanceof SJAccept)
		{
			SJAccept sja = (SJAccept) sjso;

			if (SJCompilerUtils.anonymousSJReturn(sja))
			{
				throw new SemanticException("Anonymous server-accept not yet done: " + sja);
			}

			String session = SJCompilerUtils.getSJReturnName(sja);

			sjContext().openSession(session, sjsoType); // Well-formedness checked by type builder.
			sjContext().advanceSession(session, sjsoType.nodeClone()); // 'begin'.

			SJCompilerUtils.debugPrintln("Opened: `" + session + "' " + sjsoType);
		}

		SJCompilerUtils.debugPrintln("Checked: `" + sjso + "' : " + sjsoType);

		return sjso;
	}

	protected Node overrideIf(If ifn) throws SemanticException // Duplication from parent class. Only way to share it would be to use super call and 'justPopped' or make it static.
	{
		ifn = ifn.cond((Expr) ifn.cond().visit(this));

		sjContext().pushIf();
		sjContext().pushIfBranch();

		Stmt consequent = ifn.consequent();

		if (consequent != null)
		{
			ifn = ifn.consequent((Stmt) consequent.visit(this));
		}

		sjContext().popBranchCaseContext();
		sjContext().pushIfBranch();

		Stmt alternative = ifn.alternative();

		if (alternative != null)
		{
			ifn = ifn.alternative((Stmt) alternative.visit(this));
		}

		sjContext().popBranchCaseContext();
		sjContext().updateContext((SJBranchContext) sjContext().pop()); // The only difference from base implementation of this routine.

		return ifn;
	}

	private SJCall checkSJCall(Node parent, SJCall sjc) throws SemanticException
	{
		if (sjc.isTargetImplicit()) // What about calls to a method of an outer class? Safe?
		{
			List<SJLocalSocket> socketArgs = SJCompilerUtils.getSJCallArgs(sjc); // Actually, the session types for these sockets are dynamic information, so can only look up in SJSessionTypeChecker.
			List<SJType> sjArgTypes = getSocketArgTypes(socketArgs); // Checks inactive session sockets are in scope.

			SJMethodInstance sjmi = sjContext().findSJMethod(sjc.name(), sjArgTypes);

			for (SJLocalSocket arg : socketArgs)
			{
				String session = arg.name();
				SJType sjType = sjContext().activeType(session);

				if (sjContext().sessionOpen(session))
				{
					/*if (/*sjType instanceof SJRecurseType &&* sjContext().currentSJMethod() instanceof SJRecursiveMethodContext) // Correct context for recursive call already checked by getSJCallArgs. // Relaxed to support non-recursion type arguments.
					{
						sjContext().completeSession(session);
					}
					else
					{
						sjContext().delegateSession(session);
					}*/

					try
					{
						if (sjContext().currentSJMethod() instanceof SJRecursiveMethodContext)
						{
							sjContext().completeSession(session);
						}
						else
						{
							sjContext().delegateSession(session);
						}
					}
					catch (SemanticException se)
					{
						sjContext().delegateSession(session);
					}
				}

				SJCompilerUtils.debugPrintln("Delegated: `" + session + "' " + sjType);
			}

			String returned = SJCompilerUtils.getSJCallReturnName(sjc); // Using AST information because currently instance objects are not built for SJCall.
			SJType sjReturnType = sjmi.sjReturnType();

			if (returned == null) // anonymousSJReturn method not ready for use yet (SJCall being built here).
			{
				if(sjReturnType != null)
				{
					if (!(parent instanceof SJReturn)) // Will be checked by the SJReturn routine.
					{
						throw new SemanticException("Anonymous session method-call return not done yet: " + sjc); // This is actually already caught by SJOperationParser.
					}
				}
			}
			else
			{
				if (sjReturnType instanceof SJBeginType)
				{
					sjContext().addSocket(sjts.SJNamedInstance(sjc.position(), sjReturnType, returned));
				}
				else
				{
					sjContext().openSession(returned, sjReturnType);
				}

				SJCompilerUtils.debugPrintln("Opened: `" + returned + "' " + sjReturnType);
			}

			sjc = SJCompilerUtils.setSJCallMethodInstance(sjc, sjmi); // SJCall has no object instance to record the session delegation/receive types.
		}
		else
		{
			throw new SemanticException("Session method-calls to external targets not done yet: " + sjc);

			/*ParsedClassType targetClass = (ParsedClassType) sjts.systemResolver().find(sjc.target().type().toString()); // Lookup external class.

			String fieldName = SJConstants.SJ_FORMAL_PREFIX + SJConstants.SJTHREAD_RUN + "_" +  sessionName; // Encoding fields are only available for external classes - translation isn't done until after this pass.

			FieldInstance fi = targetClass.fieldNamed(fieldName));

			// Then compare the active session types of session arguments with the (encoded) session types for the formal parameters to the target method.*/
		}

		return sjc;
	}

	private SJReturn checkSJReturn(SJReturn sjr) throws SemanticException
	{
		SJMethodContext sjme = sjContext().currentSJMethod();

		Expr expr = sjr.expr();
		SJType expected = sjme.sjReturnType();

		String session = null;
		SJType implemented = null;

		if (expr instanceof SJLocalSocket)
		{
			session = SJCompilerUtils.getSJNamedName(sjr); // Could use type instead of AST information.

			if (sjContext().sessionOpen(session)) // Maybe this check should be incorporated into a context operation.
			{
				implemented = sjContext().activeType(session);

				if (sjme instanceof SJRecursiveMethodContext)
				{
					SJMap sjFormals = sjme.sjFormals();
					SJType sjType = sjFormals.get(session).child();

					if (sjType != null)
					{
						implemented = (implemented == null) ? sjType : implemented.child(sjType);
					}

					for (String fname : sjFormals.keySet())
					{
						if (!fname.equals(session))
						{
							SJType sjt = sjFormals.get(fname);

							if (sjt instanceof SJRecursionType && sjt.child() != null) // Relaxed to support non-recursion type arguments.

							throw new SemanticException("Illegal tail for argument type: " + fname);
						}
					}
				}
			}
			else
			{
				throw new SemanticException("Currently cannot return `begin' typed sockets: " + sjr);

				//implemented = sjContext().socketType(session); // Although session-try translation puts in a finally close, this should still be necessarily done before the other side can open the session (close on an unopen session should do nothing). It doesn't matter if the socket being returned is registered or not (even if it is registered, because it is being returned, it will not be closed by the returner), if the other side wants to use it is has to be registered there. // FIXME: problem is the session may actually have been implemented before the return statement (the return is outside, after the session-try). This means the close thread could interfere with actions by the calling context to which the socket is returned (that's why session sockets currently cannot be reused in general).
			}
		}
		else //if (expr instanceof SJCall)
		{
			SJCall sjc = (SJCall) sjr.expr();

			implemented = SJCompilerUtils.getSJCallMethodInstance(sjc).sjReturnType();
		}

		if (implemented == null || !implemented.isSubtype(expected))
		{
			throw new SemanticException("Expected return type `" + expected.nodeToString() + "', not: " + implemented);
		}

		if (session != null && !(implemented instanceof SJBeginType))
		{
			if (sjme instanceof SJRecursiveMethodContext)
			{
				sjContext().completeSession(session);
			}
			else
			{
				sjContext().delegateSession(session);
			}

			SJCompilerUtils.debugPrintln("Delegated: `" + session + "' " + implemented);
		}

		implemented = sjts.SJSendType(implemented);

		return (SJReturn) SJCompilerUtils.setSJNamedType(sjr, implemented);
	}

	private SJSpawn checkSJSpawn(SJSpawn sjs) throws SemanticException // Similar to, but simpler than, checkSJCall without the return type checking.
	{
		List<Receiver> socketArgs = SJCompilerUtils.getSJSpawnArgs(sjs);

		SJThreadInstance sjti = null;

		try
		{
			sjti = sjContext().findSJThread(((ClassType) ((New) sjs.target()).type()).fullName());
		}
		catch (SemanticException se)
		{
			throw new SemanticException("SJThread not found (co-source and external classes not done yet): " + sjs);
		}

		Iterator<SJType> runTypes = sjti.runTypes().iterator();

		for (Receiver arg : socketArgs)
		{
			if (!runTypes.hasNext())
			{
				throw new SemanticException("Mismatch between arity of spawn and the target: " + sjs);
			}

			SJType expected = runTypes.next();

			String session = ((SJLocalSocket) arg).name();
			SJType implemented = null;

			if (!sjContext().sessionOpen(session)) // Doesn't matter, if session inactive, whether the socket has been registered with session-try or not? However, this means that when an unregistered socket is passed to a SJThread, it is still automatically registered in the run method (this is ok because the difference because SJCall and SJSpawn is that the latter does not need to propagate exceptions back).
			{
				implemented = sjContext().socketType(session);
			}
			else
			{
				implemented = sjContext().activeType(session);
			}

			if (implemented == null || !expected.isSubtype(implemented))
			{
				throw new SemanticException("Expected `" + expected + "' argument, not: " + implemented);
			}

			if (!(implemented instanceof SJBeginType))
			{
				sjContext().delegateSession(session);
			}

			SJCompilerUtils.debugPrintln("Delegated: `" + session + "' " + implemented);
		}

		if (runTypes.hasNext())
		{
			throw new SemanticException("Mismatch between arity of spawn and the target: " + sjs);
		}

		return SJCompilerUtils.setSJSpawnThreadInstance(sjs, sjti);
	}

	private List<SJType> getSocketArgTypes(List<SJLocalSocket> socketArgs) throws SemanticException
	{
		List<SJType> sjArgTypes = new ArrayList<SJType>();

		for (SJLocalSocket arg : socketArgs)
		{
			String session = arg.name();

			SJType sjType;

			if (!sjContext().sessionOpen(session))
			{
				if (!sjContext().socketInScope(session))
				{
					throw new SemanticException("Socket not registered with session-try: " + session);
				}

				sjType = sjContext().socketType(session);
			}
			else
			{
				sjType = sjContext().activeType(session); // Not expectedType: can "delegate" within nested context.

				if (sjType == null)
				{
					throw new SemanticException("Cannot pass a completed session: " + session);
				}
				else if (sjType instanceof SJRecurseType)
				{
					SJContextElement sjme = sjContext().currentSJMethod();

					if (!(sjme instanceof SJRecursiveMethodContext))
					{
						throw new SemanticException("Illegal context for a recursive session method call: " + session);
					}

					SJRecursiveMethodContext rmc = (SJRecursiveMethodContext)	sjme;
					SJLabel lab = ((SJRecurseType) sjType).label(); // Target method.

					if (lab.labelValue().equals(rmc.name())) // Non-mutual ("self") recursion.
					{
						sjType = rmc.sjFormals().get(session);
					}
					else
					{
						sjType = sjContext().lookupRecursionArg(lab, sjArgTypes.size()); // Hacky.
					}
				}
			}

			sjArgTypes.add(sjType);
		}

		return sjArgTypes;
	}
}
