/*
 * This file is part of SessionJ.
 *
 * Copyright (C) Raymond Hu, Imperial College London.
 *
 */

package sj.visit;

import java.util.*;

import polyglot.ast.*;
import polyglot.frontend.Job;
import polyglot.types.*;
import polyglot.util.*;
import polyglot.visit.*;

import sj.SJConstants;
import sj.ast.*;
import sj.ast.proceduredecls.*;
import sj.ast.sessformals.*;
//import sj.types.*;

/**
 * SJSpawn calls translated by (SJSessionOperationParser and) SJHigherOrderTranslator. Must come before SJSessionTryTranslator.
 */
public class SJThreadTranslator extends ErrorHandlingVisitor // ContextVisitor pass broken because types not built for new nodes created in preceding translation passes. Specifically when LocalDecls are inserted (and problems trying to build types for them).
{
	//private SJTypeSystem sjts = (SJTypeSystem) typeSystem();
	private SJNodeFactory sjnf = (SJNodeFactory) nodeFactory();

	private Stack<ClassDecl> classDecls = new Stack<ClassDecl>(); // Because can't look up currentClass using ContextVisitor.
	private Stack<List<SJFormal>> runArgs = new Stack<List<SJFormal>>(); // Seems we have to use an indirect collection to store this information and retrieve it at a later point.

	public SJThreadTranslator(Job job, TypeSystem ts, NodeFactory nf)
	{
		super(job, ts, nf);
	}

	protected NodeVisitor enterCall(Node parent, Node n) throws SemanticException
	{
		if (n instanceof ClassDecl)
		{
			classDecls.push((ClassDecl) n);
		}

		return this;
	}

	protected Node leaveCall(Node old, Node n, NodeVisitor v) throws SemanticException
	{
		if (n instanceof SJMethodDecl)
		{
			SJMethodDecl sjmd = (SJMethodDecl) n;

			//if (sjmd.name().equals(SJConstants.SJTHREAD_RUN) && context().currentClass().isSubtype(SJConstants.SJTHREAD_TYPE))
			if (sjmd.name().equals(SJConstants.SJTHREAD_RUN) && classDecls.peek().type().isSubtype(SJConstants.SJTHREAD_TYPE))
			{
				n = translateSJRunMethod(sjmd); // Pushes runArgs.
			}
		}
		else if (n instanceof ClassDecl)
		{
			if (((ClassDecl) n).type().isSubtype(SJConstants.SJTHREAD_TYPE))
			{
				n = translateSJThread((ClassDecl) n); // Adds fields and translates the constructors.

				runArgs.pop();
			}

			classDecls.pop();
		}

		return n;
	}

	private SJMethodDecl translateSJRunMethod(SJMethodDecl sjmd) throws SemanticException
	{
		/*if (runArgs.size() != 0) // Breaks nested SJThread classes. But this check should already have been done in an earlier pass (actually, currently SJContextVisitor).
		{
			throw new SemanticException("SJThreads can only have one run method: " + sjmd);
		}*/

		List<SJFormal> sjFormals = new LinkedList<SJFormal>();
		List<Receiver> sockets = new LinkedList<Receiver>();

		for (Object formal : sjmd.formals())
		{
			if (!(formal instanceof SJFormal))
			{
				throw new SemanticException("Only SJFormals allowed for SJThread run method: " + sjmd);
			}

			SJFormal sjFormal = (SJFormal) formal;

			sjFormals.add(sjFormal); // Just to cast to correct type.
			sockets.add(sjnf.SJLocalSocket(sjFormal.position(), sjFormal.id()));
		}

		sjmd = (SJMethodDecl) sjmd.formals(new LinkedList<Formal>());
		sjmd = (SJMethodDecl) sjmd.throwTypes(new LinkedList<TypeNode>());

		Position pos = sjmd.position();

		List<Catch> catchBlocks = new LinkedList<Catch>();

		catchBlocks.add(sjnf.Catch(pos, sjnf.Formal(pos, Flags.NONE, sjnf.CanonicalTypeNode(pos, SJConstants.JAVA_EXCEPTION_TYPE), sjnf.Id(pos, SJConstants.SJ_THREAD_RUNEXCFORMAL)), sjnf.Block(pos)));

		List<Stmt> finallyStmts = new LinkedList<Stmt>();

		//Id close = sjnf.Id(pos, SJConstants.SJSOCKET_CLOSE);

		//finallyStmts.add(sjnf.Throw(pos, sjnf.New(pos, sjnf.CanonicalTypeNode(pos, SJConstants.SJRUNTIMEEXCEPTION_TYPE), new LinkedList()))); // FIXME: broken when threads are not associated to a thread group? // Fix by changing socket status when grouped and checking this status at runtime (i.e. throw if grouped).

		SJSessionTry tryn = sjnf.SJSessionTry(pos, sjmd.body(), catchBlocks, sjnf.Block(pos, finallyStmts), sockets);

		sjmd = (SJMethodDecl) sjmd.body(sjnf.Block(pos, tryn));

		runArgs.push(sjFormals);

		return sjmd;
	}

	private ClassDecl translateSJThread(ClassDecl cd) throws SemanticException
	{
		if (runArgs.size() == 0)
		{
			throw new SemanticException("SJThread must have a run method: " + cd);
		}

		ClassBody body = cd.body();

		List members = body.members();
		List<ClassMember> translated = new LinkedList<ClassMember>();

		Position pos = ((ClassMember) members.get(0)).position();
		TypeNode socketType = sjnf.CanonicalTypeNode(pos, SJConstants.SJSOCKET_TYPE);

		for (SJFormal sjFormal : runArgs.peek())
		{
			translated.add(sjnf.FieldDecl(pos, Flags.PRIVATE, socketType, sjnf.Id(pos, sjFormal.name()), sjnf.NullLit(pos)));
		}

		for (Object member : members)
		{
			if (member instanceof ConstructorDecl)
			{
				ConstructorDecl cmemb = (ConstructorDecl) member;

				List<Formal> cformals = new LinkedList<Formal>();

				cformals.addAll((List<Formal>) cmemb.formals());

				for (SJFormal sjFormal : runArgs.peek())
				{
					cformals.add(sjnf.Formal(pos, Flags.NONE, socketType, sjFormal.id()));
				}

				Block cbody = cmemb.body();

				for (SJFormal sjFormal : runArgs.peek())
				{
					Id id = sjFormal.id();

					FieldAssign fa = sjnf.FieldAssign(pos, sjnf.Field(pos, sjnf.Special(pos, Special.THIS), id), Assign.ASSIGN, sjnf.Local(pos, id));

					cbody = cbody.append(sjnf.Eval(pos, fa));
				}

				translated.add((ClassMember) cmemb.formals(cformals).body(cbody));
			}
			else
			{
				translated.add((ClassMember) member);
			}
		}

		return cd.body(body.members(translated));
	}
}
