/*
 * This file is part of SessionJ.
 *
 * Copyright (C) Raymond Hu, Imperial College London.
 *
 */

package sj.visit.contexts;

import java.util.*;

import polyglot.ast.*;
import polyglot.types.*;

import sj.SJConstants;
import sj.ast.proceduredecls.*;
import sj.ast.sessformals.*;
import sj.types.SJTypeSystem;
import sj.types.typeobjects.sesstypes.*;
import sj.types.typeobjects.typeableinstances.*;
import sj.visit.contexts.sesscontexts.*;
import sj.util.*;

/**
 * Protocol and channel (and socket) type information is "static", can search through type stack (i.e. cross contexts) to look it up. (Whereas session type information is "dynamic", changes from context to context as session operations are performed - this information is copied and updated across contexts.)
 */
public class SJContext_c implements SJContext
{
	private Stack<SJContextElement> contexts = new Stack<SJContextElement>(); // Maybe make a factory for the context elements.
	private SJContextElement parent;

	private SJTypeSystem sjts = null; // Now shouldn't be needed? TIs passed in.

	public SJContext_c(SJTypeSystem sjts)
	{
		this.sjts = sjts;
	}

	//private SJContextElement currentContext()
	public SJContextElement currentContext()
	{
		return contexts.peek();
	}

	public SJMethodContext currentSJMethod() throws SemanticException
	{
		for (int i = contexts.size() - 1; i >= 0; i--)
		{
			SJContextElement sjce = contexts.get(i);

			if (sjce instanceof SJMethodContext)
			{
				return (SJMethodContext) sjce;
			}
		}

		throw new SemanticException("Not in a SJMethodContext.");
	}

	public SJContextElement parentContext()
	{
		return parent;
	}

	public void addProtocol(SJNamedInstance sjni)
	{
		currentContext().addProtocol(sjni); // Also register the type information with container (for field-protocols)? Any benefit to do so? // Maybe for translation work.
	}

	public void addChannel(SJNamedInstance sjni)
	{
		currentContext().addChannel(sjni);
	}

	public void addSocket(SJNamedInstance sjni)
	{
		currentContext().addSocket(sjni);
	}

	public void addServer(SJNamedInstance sjni)
	{
		currentContext().addServer(sjni);
	}

	public SJNamedInstance findProtocol(String name) throws SemanticException
	{
		for (int i = contexts.size() - 1; i >= 0; i--) // This back-search loop structure occurs many times in this class.
		{
			SJContextElement sjce = contexts.get(i);

			if (sjce.hasProtocol(name))
			{
				return sjce.getProtocol(name);
			}
		}

		throw new SemanticException("Protocol not in context: " + name);
	}

	public SJNamedInstance findChannel(String name) throws SemanticException
	{
		for (int i = contexts.size() - 1; i >= 0; i--)
		{
			SJContextElement sjce = contexts.get(i);

			if (sjce.hasChannel(name))
			{
				return sjce.getChannel(name);
			}
		}

		throw new SemanticException("Channel not in context: " + name);
	}

	public SJNamedInstance findSocket(String name) throws SemanticException
	{
		for (int i = contexts.size() - 1; i >= 0; i--)
		{
			SJContextElement sjce = contexts.get(i);

			if (sjce instanceof SJClassContext) // Channel and server cases may also need this.
			{
				break;
			}
			else if (sjce.hasSocket(name))
			{
				return sjce.getSocket(name);
			}
		}

		throw new SemanticException("Socket not in context: " + name);
	}

	public SJNamedInstance findServer(String name) throws SemanticException
	{
		for (int i = contexts.size() - 1; i >= 0; i--)
		{
			SJContextElement sjce = contexts.get(i);

			if (sjce.hasServer(name))
			{
				return sjce.getServer(name);
			}
		}

		throw new SemanticException("Server not in context: " + name);
	}

	public SJMethodInstance findSJMethod(String name, List<SJType> argTypes) throws SemanticException
	{
		for (int i = contexts.size() - 1; i >= 0; i--)
		{
			SJContextElement sjce = contexts.get(i);

			if (sjce instanceof SJClassContext)
			{
				SJMethodInstance sjmi = ((SJClassContext) sjce).findSJMethod(name, argTypes);

				if (sjmi != null)
				{
					return sjmi;
				}

				// Safe to continue searching through outer classes?
			}
		}

		throw new SemanticException("Session-method not found: " + name + " " + argTypes);
	}

	public SJThreadInstance findSJThread(String name) throws SemanticException
	{
		for (int i = contexts.size() - 1; i >= 0; i--)
		{
			SJContextElement sjce = contexts.get(i);

			if (sjce instanceof SJClassContext)
			{
				SJThreadInstance sjni = ((SJClassContext) sjce).findSJThread(name);

				if (sjni != null)
				{
					return sjni;
				}
			}
		}

		throw new SemanticException("Session-thread not found: " + name);
	}

	public SJType lookupRecursionArg(SJLabel lab, int argpos) throws SemanticException
	{
		for (int i = contexts.size() - 1; i >= 0; i--)
		{
			SJContextElement sjce = contexts.get(i);

			if (sjce instanceof SJClassContext) // Have to manually search through each method to see if there is a match with the SJCall being checked. Can't find the method directly, because don't know the formal types for the recursive arguments (that's what we need to find). Only works (?) because overloading (including method arity) not allowed for session methods - so the method name is unique.
			{
				SJClassContext sjcc = (SJClassContext) sjce;

				for (SJMethodInstance sjmi : sjcc.sjMethods())
				{
					if (sjmi.name().equals(lab.labelValue()))
					{
						//return sjmi.sjFormals().get(session); // Cannot use argument name to look up a formal, of course.
						return sjmi.sjFormals().get(argpos);
					}

					// Search has failed at this point (sessrec methods cannot be overloaded)?
				}

				// And here (safe or not to allow external access from inner classes)?
			}
		}

		throw new SemanticException("Invalid recursive session-method call for: " + lab);
	}

	public SJType protocolType(String name) throws SemanticException
	{
		return findProtocol(name).sjType(); // Clone?
	}

	public SJType channelType(String name) throws SemanticException
	{
		return findChannel(name).sjType();
	}

	public SJType socketType(String name) throws SemanticException
	{
		return findSocket(name).sjType();
	}

	public SJType serverType(String name) throws SemanticException
	{
		return findServer(name).sjType();
	}

	public boolean socketInScope(String name)
	{
		for (int i = contexts.size() - 1; i >= 0; i--)
		{
			SJContextElement ce = contexts.get(i);

			if (ce instanceof SJClassContext)
			{
				return false;
			}
			else if (ce instanceof SJSessionTryContext) // Only need to check innermost session-try.
			{
				return ((SJSessionTryContext) ce).socketSpecified(name);
			}
		}

		throw new RuntimeException("Shouldn't have got here: " + name);
	}

	public Set<String> sessionsInScope() throws SemanticException
	{
		SJContextElement sjce = currentContext();

		if (!(sjce instanceof SJSessionContext))
		{
			throw new SemanticException("Current context is not a session context.");
		}

		return ((SJSessionContext) sjce).sessionsInScope();
	}

	public boolean sessionOpen(String session)
	{
		SJContextElement sjce = currentContext();

		if (!(sjce instanceof SJSessionContext))
		{
			return false;
		}

		return (((SJSessionContext) sjce).sessionOpen(session));
	}

	public boolean sessionInScope(String session) throws SemanticException
	{
		return sessionsInScope().contains(session); // Or can use sessionInScope directly.
	}

	public SJType activeType(String session) throws SemanticException
	{
		SJContextElement sjce = currentContext();

		if (!(sjce instanceof SJSessionContext)) // Session doesn't have to be in scope.
		{
			throw new SemanticException("Current context is not a session context.");
		}

		return ((SJSessionContext) sjce).activeType(session);
	}

	public SJType implementedType(String session) throws SemanticException
	{
		SJContextElement sjce = currentContext();

		if (!(sjce instanceof SJSessionContext)) // Session doesn't have to be in scope.
		{
			throw new SemanticException("Current context is not a session context.");
		}

		return ((SJSessionContext) sjce).implementedType(session);
	}

	public SJType expectedType(String session) throws SemanticException
	{
		SJSessionContext sjsc = (SJSessionContext) currentContext();

		if (sessionInScope(session))
		{
			return sjsc.activeType(session);
		}
		else if (sjsc instanceof SJBranchCaseContext) // "Safe" to get expected type (and advance) for non-scope sessions within branch-cases because there is an outer branch context that makes ensures correctness across all cases.
		{
			if (sjsc.sessionOpen(session))
			{
				return sjsc.activeType(session);
			}
		}

		throw new SemanticException("Session not in scope: " + session);
	}

	public void activeType(String session, SJType sjType) throws SemanticException
	{
		SJContextElement sjce = currentContext();

		if (!(sjce instanceof SJSessionContext)) // Session doesn't have to be in scope.
		{
			throw new SemanticException("Current context is not a session context.");
		}

		((SJSessionContext) sjce).activeType(session, sjType);
	}

	public void implementedType(String session, SJType sjType) throws SemanticException
	{
		SJContextElement sjce = currentContext();

		if (!(sjce instanceof SJSessionContext)) // Session doesn't have to be in scope.
		{
			throw new SemanticException("Current context is not a session context.");
		}

		((SJSessionContext) sjce).implementedType(session, sjType);
	}

	public void openSession(String session, SJType sjType) throws SemanticException
	{
		if (sessionOpen(session))
		{
			throw new SemanticException("Session already open: " + session);
		}

		if (!socketInScope(session))
		{
			throw new SemanticException("Session socket not in valid context: " + session);
		}

		((SJSessionContext) currentContext()).openSession(session, sjType);
	}

	public void advanceSession(String session, SJType sjType) throws SemanticException
	{
		SJSessionContext sjsc = (SJSessionContext) currentContext();

		SJType implemented = implementedType(session);

		if (implemented == null)
		{
			implemented = sjType;
		}
		else
		{
			implemented = implemented.append(sjType);
		}

		sjsc.activeType(session, activeType(session).child());
		sjsc.implementedType(session, implemented);
	}

	// Uses advanceSession to advance a set of session implementations (in particular, non-scope sessions advanced in an inner context) with handling for branch cases.
	public void updateContext(SJSessionContext sce) throws SemanticException
	{
		SJContextElement sjce = currentContext();

		if (!(sjce instanceof SJSessionContext))
		{
			// Don't need to do anything: all sessions must have been completed already.
		}
		else
		{
			if (sce instanceof SJBranchContext) // The right place to do this?
			{
				sce = checkBranchCases((SJBranchContext) sce);
			}

			for (String s : sce.implementedSessions())
			{
				SJType implementedType = sce.implementedType(s);

				if (((SJSessionContext) currentContext()).sessionOpen(s)) // Ignore nested (and completed) session implementations.
				{
					for (; implementedType != null; implementedType = implementedType.child())
					{
						advanceSession(s, implementedType.nodeClone());
					}
				}
			}
		}
	}

	public SJBranchContext checkBranchCases(SJBranchContext bc) throws SemanticException
	{
		Iterator<SJBranchCaseContext> i = bc.branchCases().iterator();

		SJBranchCaseContext res = i.next(); // The first branch case.

		for (; i.hasNext(); )
		{
			SJBranchCaseContext bcc = i.next();

			for (String s : bcc.activeSessions()) // All open sessions (including outer scopes).
			{
				if (!res.sessionOpen(s)) // 's' is nested within bcc.
				{
					if (bcc.activeType(s) != null) // Check session completed. (Unnecessary?)
					{
						throw new SemanticException("Bad branch case: nested session not completed.");
					}
				}
				else
				{
					SJType existing = res.implementedType(s);
					SJType next = bcc.implementedType(s);

					if (existing == null)
					{
						if (next != null)
						{
							throw new SemanticException("Bad branch case: " + next);
						}
					}
					else
					{
						if (next == null)
						{
							throw new SemanticException("Bad branch case: ");
						}

						res.implementedType(s, existing.subsume(next));
					}
				}
			}

			for (String s : res.activeSessions())
			{
				if (!bcc.sessionOpen(s)) // (Unnecessary?)
				{
					if (res.activeType(s) != null)
					{
						throw new SemanticException("Bad branch case: nested session not completed");
					}
				}
			}
		}

		for (String s : res.implementedSessions())
		{
			SJType sjType = res.implementedType(s);

			bc.implementedType(s, sjType);

			SJCompilerUtils.debugPrintln("Branch subsumed: " + sjType);
		}

		return bc;
	}

	public void completeSession(String session) throws SemanticException
	{
		if (!sessionOpen(session))
		{
			throw new SemanticException("Cannot complete an unopen session: " + session);
		}

		for (SJType active = activeType(session); active != null; active = activeType(session))
		{
			advanceSession(session, active.nodeClone());
		}
	}

	public void delegateSession(String session) throws SemanticException
	{
		if (!sessionOpen(session)) // Delegation of unopen sessions (inactive sockets) not allowed.
		{
			throw new SemanticException("Cannot complete an unopen session: " + session);
		}

		boolean illegalContext = false;

		try
		{
			if (currentSJMethod() instanceof SJRecursiveMethodContext)
			{
				illegalContext = true;
			}
		}
		catch (SemanticException se) { }

		if (illegalContext)
		{
			throw new SemanticException("Illegal context for delegation: " + session);
		}

		for (int i = contexts.size() - 1; i >= 0; i--)
		{
			SJContextElement sjce = contexts.get(i);

			if (sjce instanceof SJSessionContext && ((SJSessionContext) sjce).sessionInScope(session))
			{
				break; // If session is registered as in-scope, then it is necessarily open.
			}
			else if (sjce instanceof SJWhileContext || sjce instanceof SJRecursionContext)
			{
				throw new SemanticException("Illegal context for delegation: " + session);
			}
			else if (i == 0) // Unnecessary? Session would be open in current context?
			{
				throw new SemanticException("Cannot complete an unopen session: " + session);
			}
		}

		boolean foo = false;

		for (int i = contexts.size() - 1; i >= 0; i--)
		{
			SJContextElement sjce = contexts.get(i);

			if (sjce instanceof SJSessionContext)
			{
				SJSessionContext sjsc = (SJSessionContext) sjce;

				if (sjsc instanceof SJSessionTryContext && ((SJSessionTryContext) sjsc).socketSpecified(session))
				{
					if (foo)
					{
						SJType active = sjsc.activeType(session);

						if (active != null && active.child() != null)
						{
							throw new SemanticException("Session type \"tails\" not done yet for delegation: " + active.child());
						}

						break;
					}
				}
				else if (sjsc instanceof SJIOBranchCaseContext && ((SJIOBranchCaseContext) sjsc).session().equals(session))
				{
					foo = true;
				}
			}
		}

		completeSession(session);
	}

	public void pushClass(ClassDecl cd) throws SemanticException // No notion of general "session-class" so all session type checking (access modifiers, overriding, ...) and building (session class/method members) currently done here.
	{
		SJClassContext sjcc = new SJClassContext_c();

		ClassBody body = cd.body();

		// classes, methods
		// private (thread) class members, no method overriding

		for (Object member : body.members())
		{
			if (member instanceof SJProcedureDecl)
			{
				/*if (member instanceof SJConstructorDecl)
				{
					throw new SemanticException("Session constructors not done yet.");
				}
				else*/ if (member instanceof SJMethodDecl) // Includes SJThread run method.
				{
					SJMethodInstance sjmi = SJCompilerUtils.getSJMethodInstance((SJMethodDecl) member);

					if (sjmi == null) // A bit hacky: session types have not been built yet.
					{
						break; // End up pushing an empty SJClassContext.
					}

					boolean methodExists = false; // No overloading allowed for session-methods.

					try // Ugly structure due to use of exceptions.
					{
						sjcc.findSJMethod(sjmi.name(), SJCompilerUtils.getSJFormalTypes(sjmi));

						methodExists = true;
					}
					catch (SemanticException se)
					{
						sjcc.addSJMethod(sjmi);
					}

					if (methodExists)
					{
						throw new SemanticException("Overloading for session-methods not done yet: " + member);
					}
				}
			}
			else if (member instanceof ClassDecl) // Later generalise for all class types.
			{
				ClassDecl cmemb = (ClassDecl) member;

				if (!cmemb.flags().isPrivate())
				{
					throw new SemanticException("Non-private class members not done yet: " + cmemb);
				}

				if (cmemb.type().isSubtype(SJConstants.SJTHREAD_TYPE))
				{
					List<SJType> runTypes = new ArrayList<SJType>(); // If had a SJClassType instance object for SJThread or all classes, then could build this information using the SJTypeBuilder and store it there - would be easier to look up here (like in the above case).

					boolean foundRun = false;

					for (Object memb : cmemb.body().members())
					{
						if (memb instanceof MethodDecl)
						{
							MethodDecl md = (MethodDecl) memb;

							if (md.name().equals(SJConstants.SJTHREAD_RUN))
							{
								if (foundRun)
								{
									throw new SemanticException("SJThreads cannot have multiple run methods: " + cmemb);
								}

								for (Object formal : md.formals())
								{
									SJNamedInstance sjni = SJCompilerUtils.getSJNamedInstance((SJFormal) formal);

									if (sjni == null)
									{
										break; // End up pushing an empty SJClassContext.
									}

									runTypes.add(sjni.sjType());
								}

								foundRun = true;
							}
						}
					}

					if (!foundRun)
					{
						throw new SemanticException("SJThreads must have a run method: " + cmemb);
					}

					sjcc.addSJThread(sjts.SJThreadInstance(((ClassType) cmemb.memberInstance()).fullName(), runTypes));
				}
			}
		}

		contexts.push(sjcc);
	}

	public void pushCode()
	{
		contexts.push(new SJCodeContext_c());
	}

	public void pushTry()
	{
		if (currentContext() instanceof SJSessionContext)
		{
			contexts.push(new SJSessionTryContext_c(new HashSet<String>(), (SJSessionContext) currentContext())); // Make a separate base TryContext element? // Just use an ordinary code context?
		}
		else
		{
			contexts.push(new SJSessionTryContext_c(new HashSet<String>()));
		}
	}

	public void pushSJSessionTry(Set<String> sockets) throws SemanticException
	{
		for (int i = contexts.size() - 1; i >= 0; i--)
		{
			SJContextElement sjce = contexts.get(i);

			if (sjce instanceof SJClassContext)
			{
				break; // Don't need to search back any further?
			}
			else if (sjce instanceof SJSessionTryContext)
			{
				for (String s : sockets)
				{
					if (((SJSessionTryContext) sjce).socketSpecified(s))
					{
						throw new SemanticException("Socket already registered to an outer scope: " + s);
					}
				}
			}
		}

		SJContextElement current = currentContext();

		if (current instanceof SJSessionContext)
		{
			contexts.push(new SJSessionTryContext_c(sockets, ((SJSessionContext) current)));
		}
		else
		{
			contexts.push(new SJSessionTryContext_c(sockets));
		}
	}

	public void pushIf()
	{
		if (currentContext() instanceof SJSessionContext) // This was moved out of the session context element constructor - bad idea?
		{
			contexts.push(new SJBranchContext_c((SJSessionContext) currentContext()));
		}
		else
		{
			contexts.push(new SJBranchContext_c());
		}
	}

	public void pushIfBranch()
	{
		if (currentContext() instanceof SJSessionContext)
		{
			contexts.push(new SJBranchCaseContext_c((SJSessionContext) currentContext()));
		}
		else
		{
			contexts.push(new SJBranchCaseContext_c());
		}
	}

	public void pushWhile()
	{
		if (currentContext() instanceof SJSessionContext)
		{
			contexts.push(new SJWhileContext_c((SJSessionContext) currentContext()));
		}
		else
		{
			contexts.push(new SJWhileContext_c());
		}
	}

	public void pushSJOutbranch(String session, SJLabel lab) throws SemanticException
	{
		SJType expected = expectedType(session); // Session type checker ensures current context is a session context.

		if (!(expected instanceof SJOutbranchType))
		{
			throw new SemanticException("Expected outbranch type, not: " + expected);
		}

		SJOutbranchType obt = (SJOutbranchType) expected;

		if (!obt.hasCase(lab))
		{
			throw new SemanticException("Invalid label: " + lab);
		}

		expected = obt.branchCase(lab);

		if(expected != null)
		{
			expected = expected.treeClone();
		}

		contexts.push(new SJIOBranchCaseContext_c(session, lab, expected, (SJSessionContext) currentContext()));
	}

	public void pushSJInbranch(String session) throws SemanticException
	{
		SJType expected = expectedType(session);

		if (!(expected instanceof SJInbranchType))
		{
			throw new SemanticException("Expected inbranch type, not: " + expected);
		}

		contexts.push(new SJInbranchContext_c(session, ((SJInbranchType) expected).nodeClone(), (SJSessionContext) currentContext()));
	}

	public void pushSJInbranchCase(SJLabel lab) throws SemanticException
	{
		String session = ((SJInbranchContext) currentContext()).session();

		SJInbranchType ibt = (SJInbranchType) activeType(session); // Not using expectedType - branch contexts have no sessions in scope.

		if (ibt == null)
		{
			throw new SemanticException("Over specification of inbranch types not supported: " + lab);
		}

		if (!ibt.hasCase(lab))
		{
			throw new SemanticException("Unspecified branch label: " + lab); // Not full branch subtpying.
		}

		SJType expected = ibt.branchCase(lab);

		if(expected != null)
		{
			expected = expected.treeClone();
		}

		contexts.push(new SJIOBranchCaseContext_c(session, lab, expected, (SJSessionContext) currentContext()));
	}

	public void pushSJOutwhile(String session) throws SemanticException
	{
		SJType expected = expectedType(session);

		if (!(expected instanceof SJOutwhileType))
		{
			throw new SemanticException("Expected outwhile type, not: " + expected);
		}

		expected = ((SJOutwhileType) expected).body();

		if(expected != null)
		{
			expected = expected.treeClone();
		}

		contexts.push(new SJIOWhileContext_c(session, expected, (SJSessionContext) currentContext()));
	}

	public void pushSJInwhile(String session) throws SemanticException
	{
		SJType expected = expectedType(session);

		if (!(expected instanceof SJInwhileType))
		{
			throw new SemanticException("Expected inwhile type, not: " + expected);
		}

		expected = ((SJInwhileType) expected).body();

		if(expected != null)
		{
			expected = expected.treeClone();
		}

		contexts.push(new SJIOWhileContext_c(session, expected, (SJSessionContext) currentContext()));
	}

	public void pushSJRecursion(String session, SJLabel lab) throws SemanticException
	{
		SJType expected = expectedType(session);

		if (!(expected instanceof SJRecursionType))
		{
			throw new SemanticException("Expected recursion type, not: " + expected);
		}

		SJRecursionType sjrt = (SJRecursionType) expected;

		if (!sjrt.label().equals(lab))
		{
			throw new SemanticException("Expected recursion label `" + sjrt.label() + "', not: " + lab);
		}

		expected = sjrt.body();

		if(expected != null)
		{
			expected = expected.treeClone();
		}

		contexts.push(new SJRecursionContext_c(session, lab, expected, (SJSessionContext) currentContext()));
	}

	public void pushSJMethodDecl(SJMethodInstance sjmi)
	{
		SJMap sjFormals = sjmi.sjFormals(); // Maybe just directly record the instance object in the context, but problem is need the "socket set" for the constructor super call.

		Set<String> sockets = new HashSet<String>();

		for (String session : sjFormals.keySet())
		{
			SJType sjType = sjFormals.get(session);

			if (sjType instanceof SJBeginType)
			{
				sockets.add(session); // Only the inactive sessions are recorded as sockets for the base session-try context. Would it be useful to record all the sessions for this? // SJCall rules ensure the passed inactive session sockets are in scope (registered with session-try) at the caller.
			}
		}

		contexts.push(new SJMethodContext_c(sjmi.name(), sockets, sjFormals, sjmi.sjReturnType()));

		for (String socket : sockets)
		{
			addSocket(sjts.SJNamedInstance(null, sjFormals.get(socket), socket)); // Hacky? No position value for these sockets (could use the formal parameter position - need to pass that information in).
		}
	}

	public void pushSJRecursiveMethodDecl(SJMethodInstance sjmi)
	{
		contexts.push(new SJRecursiveMethodContext_c(sjmi.name(), sjmi.sjFormals(), sjmi.sjReturnType()));
	}

	public SJContextElement pop()
	{
		this.parent = contexts.pop();

		return parentContext();
	}

	public SJContextElement popBranchCaseContext()
	{
		SJBranchCaseContext sjbcc = (SJBranchCaseContext) pop();

		((SJBranchContext) currentContext()).addBranchCase(sjbcc);

		return sjbcc;
	}
}
