/*
 * This file is part of SessionJ.
 *
 * Copyright (C) Raymond Hu, Imperial College London.
 *
 */

package sj.runtime.net;

import java.io.*;
import java.net.*;
import java.util.*;
import java.util.Map.Entry;

import sj.types.typeobjects.sesstypes.*;
import sj.util.*;

import sj.runtime.*;
import sj.runtime.util.*;

public final class SJFSocket extends SJSocket
{
	protected static final boolean RUNTIME_STATE = false;
	protected static final boolean CHECK_DUALITY = true;

	//BufferedReader br = new BufferedReader(new InputStreamReader(System.in));
	//try { br.readLine(); } catch (IOException ioe) { }

	private static final byte SJ_FIN = 0; // Need to distinguish success and failure FIN.
	private static final byte SJ_OBJECT = 1;
	private static final byte SJ_BOOLEAN = 2;
	private static final byte SJ_INT = 3;
	private static final byte SJ_BYTE = 4;

	private static final SJFIN FIN = new SJFIN();

	private final SJType protocolType; // Is this useful for stateless sockets?
	protected SJSocketState state; // Move to base class? But FSocket doesn't really need runtime type tracking and it could be optimised out.

	private ObjectOutputStream oos;
	private ObjectInputStream ois;

	// For delegation.
	//private static final SJDSAck DS_ACK = new SJDSAck(); // Using message type for protocol control, but maybe faster to use e.g. int values (can SJ_FIN, ACK, etc. be replaced solely by flags?).
	private static final byte SJ_START_DEL = 10;
	private static final byte SJ_DEL_SIGNAL = 11;
	private static final byte SJ_DEL_ACK = 12;

	// For simultaneous delegation.
	private List<Object> lostMessages = new LinkedList<Object>();

	public static final SJFSocket create(SJServerAddress channel)
	{
		return new SJFSocket(channel);
	}

	private SJFSocket(SJServerAddress channel)
	{
		setChannel(channel);

		this.protocolType = decodeSJType(channel.getProtocol()); // All message types specified in our protocol must be locally available (although the remote class loader has been created, codebase and other parameters have not been set yet).

		if (RUNTIME_STATE)
		{
			this.state = new SJSocketState(sjTypeSystem(), protocolType); // State object will handle cloning.
		}
	}

	protected SJFSocket(SJType sjType, SJProtocol encoded, Socket socket)
	{
		setChannel(SJServerAddress.create(encoded, (InetSocketAddress) socket.getRemoteSocketAddress()));

		setSocket(socket); // TCP_NO_DELAY already set by server (convenient exception handling).

		if (!RUNTIME_STATE)
		{
			this.protocolType = sjType; // Needed for duality check.
		}
		else
		{
			this.protocolType = sjType.treeClone();
			this.state = new SJSocketState(sjTypeSystem(), protocolType);
		}
	}

	public final void request() throws UnknownHostException, SJIOException, SJIncompatibleSessionException // request should be synchronised with any spawned close threads, i.e. cannot comment a new run until the previous has closed. Then session sockets can be reused (and `begin' typed sockets can be returned by method calls).
	{
		checkClassLoadingParameters();

		connect(getChannel().targetAddress());
		init(); // Throws an SJException if any initiation problems. // Verify compatible socket implementations?

		isActive(true);
		setOriginalRequestor(IS_ORIGINAL_REQUESTOR);

		if (RUNTIME_STATE)
		{
			state.request();

			//SJRuntimeUtils.debugPrintln("Current state: " + state.currentState());
		}
	}

	private final void connect(InetSocketAddress addr) throws SJIOException
	{
		try
		{
			Socket socket = new Socket(addr.getAddress(), addr.getPort());

			socket.setTcpNoDelay(TCP_NO_DELAY);

			setSocket(socket);
			requestIOStreams();

			//SJRuntimeUtils.debugPrintln("Connected on port " + socket.getLocalPort() + " to: " + socket.getInetAddress().getHostName() + ":" + socket.getPort());
		}
		catch (IOException ioe)
		{
			throw new SJIOException(ioe);
		}
	}

	private final void requestIOStreams() throws SJIOException
	{
		try
		{
			Socket s = getSocket();

			this.oos = new ObjectOutputStream(s.getOutputStream());

			if (remoteClassLoadingEnabled())
			{
				SJRemoteClassLoader sjrcl = new SJRemoteClassLoader(this.getClass().getClassLoader(), this);

				this.ois = new SJObjectInputStream(s.getInputStream(), sjrcl);

				setRemoteClassLoader(sjrcl);
			}
			else
			{
				this.ois = new ObjectInputStream(s.getInputStream());
			}
		}
		catch (IOException ioe)
		{
			throw new SJIOException(ioe);
		}
	}

	protected final void acceptIOStreams() throws SJIOException
	{
		try
		{
			Socket s = getSocket();

			if (remoteClassLoadingEnabled())
			{
				SJRemoteClassLoader sjrcl = new SJRemoteClassLoader(this.getClass().getClassLoader(), this);

				this.ois = new SJObjectInputStream(s.getInputStream(), sjrcl);

				setRemoteClassLoader(sjrcl);
			}
			else
			{
				this.ois = new ObjectInputStream(s.getInputStream());
			}

			this.oos = new ObjectOutputStream(s.getOutputStream());
		}
		catch (IOException ioe)
		{
			throw new SJIOException(ioe);
		}
	}

	public final void sendSession(SJSocket socket, String encoded) throws SJIOException
	{
		if (!(socket instanceof SJFSocket))
		{
			throw new SJIOException("Cannot delegate a session socket of different implementation: " + socket.getClass());
		}

		SJFSocket fsocket = (SJFSocket) socket; // fsocket is active, guaranteed by typing.

		try
		{
			fsocket.isActive(false);

			SJStartDelegation sjsd = new SJStartDelegation(fsocket.isOriginalRequestor(), fsocket.getSocketParameters());

			writeByte(SJ_START_DEL);
			writeObject(sjsd);

			int port = readIntWithDelegationCheck();

			SJDelegationSignal ourds = null;

			if (!RUNTIME_STATE)
			{
				ourds = new SJDelegationSignal(getTargetAddress().getHostName(), port);
			}
			else
			{
				ourds = new SJDelegationSignal(getTargetAddress().getHostName(), port, encoded); // Better to use actual dynamically recorded state?
			}

			fsocket.writeByte(SJ_DEL_SIGNAL);
			fsocket.writeObject(ourds);

			while (true)
			{
				byte flag = fsocket.readByte();

				if (flag == SJ_DEL_ACK) // Delegation cases 1-4.
				{
					//SJRuntimeUtils.debugPrintln("Session-send forwarding: DEL_ACK");

					writeByteAndFlush(SJ_DEL_ACK);

					break;
				}
				else if (flag == SJ_FIN) // FIXME: need to distinguish success and failure FIN, currently don't know if passive party has failed or actually finished (currently assuming the latter).
				{
					//SJRuntimeUtils.debugPrintln("Session-send forwarding: FIN");

					writeByteAndFlush(SJ_FIN);

					break;
				}
				else if (flag == SJ_START_DEL)
				{
					//SJRuntimeUtils.debugPrintln("Session-send forwarding: START_DEL");

					writeByte(SJ_START_DEL);
					writeObject(fsocket.readObjectNoCNFE());

					break;
				}
				else if (flag == SJ_DEL_SIGNAL)
				{
					//SJRuntimeUtils.debugPrintln("Session-send forwarding: DEL_SIGNAL");

					writeByte(SJ_DEL_SIGNAL);
					writeObject(fsocket.readObjectNoCNFE());

					break;
				}
				else // The lost messages.
				{
					if (flag == SJ_OBJECT)
					{
						Object obj = fsocket.readObjectNoCNFE();

						//SJRuntimeUtils.debugPrintln("Session-send forwarding: " + obj);

						writeByte(SJ_OBJECT);
						writeObject(obj); // Reserialized after we just derialized, maybe optimise.
					}
					else if (flag == SJ_BOOLEAN)
					{
						//SJRuntimeUtils.debugPrintln("Session-send forwarding: boolean");

						writeByte(SJ_BOOLEAN);
						writeBoolean(fsocket.readBoolean()); // Flushes.
					}
					else if (flag == SJ_INT)
					{
						//SJRuntimeUtils.debugPrintln("Session-send forwarding: int");

						writeByte(SJ_INT);
						writeInt(fsocket.readInt()); // Flushes.
					}
					else if (flag == SJ_BYTE)
					{
						//SJRuntimeUtils.debugPrintln("Session-send forwarding: byte");

						writeByte(SJ_BYTE);
						writeByteAndFlush(fsocket.readByte());
					}
					else
					{
						throw new SJIOException("Unrecognised message flag: " + flag);
					}
				}
			}

			if (RUNTIME_STATE)
			{
				state.sendSession(fsocket.expectedType());

				//SJRuntimeUtils.debugPrintln("Current state: " + state.currentState());
			}
		}
		finally
		{
			try
			{
				//SJRuntimeUtils.debugPrintln("Closing (" + fsocket.getLocalAddress().getPort() + ")" + (RUNTIME_STATE ? (": " + state.currentState()) : ""));

				if (RUNTIME_STATE)
				{
					fsocket.state.close();
				}

				fsocket.shutdown();
			}
			catch (SJIOException sjioe)
			{
				// Should this exception be supressed?
			}
		}
	}

	public final SJSocket receiveSession(String encoded) throws SJIOException // Need to prevent receiving an endpoint to a session for which we already own the other endpoint? But may be tricky dealing with local threads.
	{
		SJProtocol protocol = new SJProtocol(encoded);

		int port = SJRuntimeUtils.findFreeTCPPort();

		ServerSocket ss = null;
		Socket s = null;

		SJStartDelegation sjsd = null;
		boolean origreq;
		boolean simdel = false;

		List<Object> lostmsgbuffer = new LinkedList<Object>();

		try
		{
			ss = new ServerSocket(port); // Could use the other s constructor here, would check compatibility (and would have to prepend a 'begin'). // Or could make seperate constructors for delegation. // No need to check class compatibility between the new peers though - already checked via the higher-order types.

			//readInt(); // SJ_START_DEL

			sjsd = (SJStartDelegation) readObjectNoCNFEWithDelCheck(); // Automatically reads buffered lost messages (because it uses readObjectWithDelegationCheck). Refactor to make this clear in the method name?

			origreq = sjsd.originalRequestor(); // Also checks if sendSession party is delegating. In forwarding, this can be moved later? And if we receive a delegation signal after writing our port, have to check if whose delegation signal it is. // Currently, origreq is not forwarded so don't need to check in LostMessages.

			//SJRuntimeUtils.debugPrintln("Start delegation report: " + sjsd.socketParameters());

			writeByte(SJ_INT);
			writeInt(port);

			//SJRuntimeUtils.debugPrintln("Waiting for reconnection on: " + port);

			//for (boolean redo = true; redo; )
			while (true)
			{
				byte flag = readByte();

				if (flag == SJ_DEL_ACK) // Delegation cases 1-4.
				{
					//SJRuntimeUtils.debugPrintln("Session-receive received forwarded: DEL_ACK");

					break; //redo = false; // break should be faster?
				}
				else if (flag == SJ_FIN)
				{
					//SJRuntimeUtils.debugPrintln("Session-receive received forwarded: FIN");

					lostmsgbuffer.add(FIN);

					s = new Socket(); // Passive party won't reconnect. // Making an unconnected s is slower than using null, but more uniform (at least can get the local address from an SJSocket instance, whether or not it's connected).

					break;
				}
				else if (flag == SJ_START_DEL)
				{
					//SJRuntimeUtils.debugPrintln("Session-receive received forwarded: START_DEL");

					lostmsgbuffer.add(readObjectNoCNFE()); // Start delegation can have variable size?

					break;
				}
				else if (flag == SJ_DEL_SIGNAL)
				{
					//SJRuntimeUtils.debugPrintln("Session-receive received forwarded: DEL_SIGNAL");

					simdel = true;

					if (origreq) // Double delegation,
					{
						SJDelegationSignal ds = (SJDelegationSignal) readObjectNoCNFE();

						String targetAddress = ds.hostAddress();
						int targetPort = ds.port();

						s = new Socket(targetAddress, targetPort); // Don't make a new SJFSocket, don't need to do initiation routines (although maybe should do eager class loading and verification).

						//SJRuntimeUtils.debugPrintln("Reconnected on port " + s.getLocalPort() + " to: " + targetAddress + ":" + targetPort);
					}

					break;
				}
				else // The lost messages.
				{
					if (flag == SJ_OBJECT)
					{
						lostmsgbuffer.add(readObjectNoCNFE());
					}
					else if (flag == SJ_BOOLEAN)
					{
						lostmsgbuffer.add(new Boolean(readBoolean()));
					}
					else if (flag == SJ_INT)
					{
						lostmsgbuffer.add(new Integer(readInt()));
					}
					else if (flag == SJ_BYTE)
					{
						lostmsgbuffer.add(new Byte(readByte()));
					}
					else
					{
						throw new SJIOException("Unrecognised message flag: " + flag);
					}

					//SJRuntimeUtils.debugPrintln("Session-receive received forwarded: " + lostmsgbuffer.get(lostmsgbuffer.size() - 1));
				}
			}

			//SJRuntimeUtils.debugPrintln("Lost messages: " + lostMessages);

			if (s == null)
			{
				s = ss.accept(); // FIXME: if other party has failed, don't want to hang forever. Maybe insert an ack, shouldn't be a big deal cost wise. Or use timeouts. (But actually alright, can't tell the difference between failure and just very slow).

				//SJRuntimeUtils.debugPrintln("Accepted reconnection on " + port  + " from: " + s.getInetAddress().getHostName() + ":"+ s.getPort());
			}

			s.setTcpNoDelay(true);
		}
		catch (IOException ioe)
		{
			throw new SJIOException(ioe);
		}
		finally
		{
			try
			{
				SJRuntimeUtils.closeServerSocket(ss);
			}
			catch (IOException ioe) { }
		}

		SJFSocket fsocket = new SJFSocket(decodeSJType(protocol), protocol, s); // When receiving delegated sessions, a s implementation can only create net of the same type.

		fsocket.setSocketParameters(sjsd.socketParameters());

		if (s.isConnected()) // No reconnection needed if a SJ_FIN was forwarded.
		{
			if (simdel && origreq) // Do as a connect...
			{
				fsocket.requestIOStreams();
			}
			else // ...or else do as the "server side".
			{
				fsocket.acceptIOStreams();
			}
		}

		fsocket.isActive(true); // Including SJ_FIN case (not connected, only reading lost messages).
		fsocket.setOriginalRequestor(origreq);

		fsocket.lostMessages = lostmsgbuffer;

		if (RUNTIME_STATE)
		{
			fsocket.state.open();

			state.receiveSession(decodeSJType(encoded)); // Would need to send delegated session type in start delegation (no real point to use the compiler embedded value).

			//SJRuntimeUtils.debugPrintln("Current state: " + state.currentState());
		}

		return fsocket;
	}

	private final void reconnectSingleDelegation(SJDelegationSignal theirds) throws SJIOException
	{
		String targetAddr = theirds.hostAddress();
		int targetPort = theirds.port();
		String encoded = theirds.targetState();
		SJType targetState = null;

		if (encoded != null) // Session delegator may not have runtime state tracking enabled.
		{
			targetState = decodeSJType(encoded);
		}

		//SJRuntimeUtils.debugPrintln("Single delegation signal report: " + targetAddr + ":" + targetPort + ", " + targetState);

		try
		{
			writeByteAndFlush(SJ_DEL_ACK);
		}
		catch (SJIOException ioe) { } // Session delegator will close delegated session immediately if we have sent SJ_FIN or start delegation: so this operation will fail, and we ignore the exception. If session delegator has genuinely failed, the session receiver will still be expecting us anyway, so safe to proceed and ignore exception because it would have been the last communication on that session anyway. And if this operation fails when the session sender is expecting it, then an exception will be raised. // Is this desirable?
		finally
		{
			try
			{
				shutdown();
			}
			catch (SJIOException ioe) { }
		}

		connect(new InetSocketAddress(targetAddr, targetPort)); // Prints a "connected" message.

		/*//SJRuntimeUtils.debugPrintln("Reconnected to: " + targetAddr + ":" + targetPort);*/
	}

	public final void sendServerAddress(SJServerAddress sjsa) throws SJIOException
	{
		writeByte(SJ_OBJECT); // Make a channel flag?
		writeObject(sjsa.targetAddress());

		if (RUNTIME_STATE)
		{
			state.sendChannel(decodeSJType(sjsa.getProtocol().encodedSJType()));

			//SJRuntimeUtils.debugPrintln("Current state: " + state.currentState());
		}
	}

	public final SJServerAddress receiveServerAddress(String encoded) throws SJIOException // Might be better to define the serialization protocol for SJServerAddress types.
	{
		InetSocketAddress isa = (InetSocketAddress) readObjectNoCNFEWithDelCheck();
		SJServerAddress channel = SJServerAddress.create(new SJProtocol(encoded), isa);

		if (RUNTIME_STATE)
		{
			state.receiveChannel(decodeSJType(encoded)); // Would need to send the channel type with the address.

			//SJRuntimeUtils.debugPrintln("Current state: " + state.currentState());
		}

		return channel;
	}

	private final class closeProtocol implements Runnable // Supports active sockets, whether connected or not.
	{
		public final void run()
		{
			boolean isConnected = isConnected();

			try
			{
				//for (boolean redo = true; redo; ) // In case of re-delegations.
				while (true)
				{
					if (isConnected) // Hacky?
					{
						writeByteAndFlush(SJ_FIN);
					}

					if (lostMessages.size() == 0) // Not possible if not connected (would contain at least the SJ_FIN that meant the socket did not need to be connected, but actually typing prevents delegation of completed sessions so would more than just the SJ_FIN).
					{
						byte flag = readByte(); // FIXME: seems to be broken for C in delegation case 4?

						if (flag == SJ_FIN)
						{
							break; //redo = false;
						}
						else if (flag == SJ_DEL_SIGNAL)
						{
							readObjectNoCNFE();

							break;
						}
						else if (flag == SJ_BOOLEAN) // After here means session has failed.
						{
							readBoolean();
						}
						else if (flag == SJ_INT)
						{
							readInt();
						}
						else if (flag == SJ_BYTE)
						{
							readByte();
						}
						else if (flag == SJ_OBJECT || flag == SJ_START_DEL)
						{
							try
							{
								readObject();
							}
							catch (ClassNotFoundException cnfe) { }
						}
						else
						{
							// Bad message flag.
						}
					}
					else
					{
						try
						{
							if (readLostMessage() instanceof SJFIN) // Else oiscard.
							{
								break;//redo = false;
							}
						}
						catch (ClassNotFoundException cnfe)
						{
							// Don't care if we can't deserialize lost messages if session has failed.
						}
					}
				}
			}
			catch (SJIOException ioe) // For forwarding socket, don't care if close protocol fails.
			{
				//ioe.printStackTrace();
			}
			finally
			{
				try
				{
					//SJRuntimeUtils.debugPrintln("Closing socket (" + (isConnected ? getLocalAddress().getPort() : UNBOUND_PORT_VALUE) + ")" + (RUNTIME_STATE ? (": " + state.currentState()) : ""));

					if (RUNTIME_STATE)
					{
						state.close();
					}

					shutdown();
				}
				catch (SJIOException ioe) { }
			}
		}
	}

	protected final void close() /*throws SJIOException*/ // Is it important for close to throw IOException?
	{
		if (isActive()) // origreq value is reset by next request (if reusing a server-accepted socket) // Although, multiple session runs using the same socket currently not permitted by compiler.
		{
			isActive(false);

			spawnCloseThread(); // An active socket may be just reading from lost message buffer, not actually connected. In this case, the close protocol will process the buffered SJ_FIN.
		}
	}

	private final void spawnCloseThread()
	{
		//SJRuntimeUtils.debugPrintln("Spawning close thread for (" + (isConnected() ? getLocalAddress().getPort() : UNBOUND_PORT_VALUE) + ")" + (RUNTIME_STATE ? (": " + state.currentState()) : ""));

		new Thread(new closeProtocol()).start();
	}

	private final void shutdown() throws SJIOException
	{
		try
		{
			SJRuntimeUtils.closeStream(oos); // Flush is broken if it doesn't go before ois close?
			SJRuntimeUtils.closeStream(ois);
			SJRuntimeUtils.closeSocket(getSocket()); // If socket is null, the streams were as well.
		}
		catch (IOException ioe) // Should unconditionally attempt all?
		{
			throw new SJIOException(ioe);
		}
	}

	public final void send(Object obj) throws SJIOException
	{
		writeByte(SJ_OBJECT);
		writeObject(obj);

		if (RUNTIME_STATE)
		{
			state.send(obj);

			//SJRuntimeUtils.debugPrintln("Current state: " + state.currentState());
		}
	}

	public final void sendBoolean(boolean v) throws SJIOException
	{
		writeByte(SJ_BOOLEAN);
		writeBoolean(v); // Flushes.

		if (RUNTIME_STATE)
		{
			state.sendBoolean(v);

			//SJRuntimeUtils.debugPrintln("Current state: " + state.currentState());
		}
	}

	public final void sendInt(int v) throws SJIOException
	{
		writeByte(SJ_INT);
		writeInt(v); // Flushes.

		if (RUNTIME_STATE)
		{
			state.sendInt(v);

			//SJRuntimeUtils.debugPrintln("Current state: " + state.currentState());
		}
	}

	public final void sendByte(byte v) throws SJIOException
	{
		writeByte(SJ_BYTE);
		writeByteAndFlush(v);

		if (RUNTIME_STATE)
		{
			state.sendByte(v);

			//SJRuntimeUtils.debugPrintln("Current state: " + state.currentState());
		}
	}

	public final Object receive() throws SJIOException, ClassNotFoundException
	{
		Object obj = readObjectWithDelegationCheck(); // Automatically reads buffered lost messages, if any. "Lost messages" might be received along with a delegated session from the party delegating the session to us. Then we should get our next expected message from the "lost messages", it's not coming from the (new) session peer.

		if (RUNTIME_STATE)
		{
			state.receive(obj);

			//SJRuntimeUtils.debugPrintln("Current state: " + state.currentState());
		}

		return obj;
	}

	public final boolean receiveBoolean() throws SJIOException
	{
		boolean v = readBooleanWithDelegationCheck();

		if (RUNTIME_STATE)
		{
			state.receiveBoolean(v);

			//SJRuntimeUtils.debugPrintln("Current state: " + state.currentState());
		}

		return v;
	}

	public final int receiveInt() throws SJIOException
	{
		int v = readIntWithDelegationCheck();

		if (RUNTIME_STATE)
		{
			state.receiveInt(v);

			//SJRuntimeUtils.debugPrintln("Current state: " + state.currentState());
		}

		return v;
	}

	public final byte receiveByte() throws SJIOException
	{
		byte v = readByteWithDelegationCheck();

		if (RUNTIME_STATE)
		{
			state.receiveByte(v);

			//SJRuntimeUtils.debugPrintln("Current state: " + state.currentState());
		}

		return v;
	}

	public final String outlabel(String lab) throws SJIOException
	{
		writeByte(SJ_OBJECT);
		writeObject(lab);

		if (RUNTIME_STATE)
		{
			state.outbranch(new SJLabel(lab)); // Some mismatch between using socket operations to coordinate protocol state. Should raise socket operations to session level? // Faster to keep labels as strings? But corresponding type objects use SJLabels.

			//SJRuntimeUtils.debugPrintln("Current state: " + state.currentState());
		}

		return lab;
	}

	public final String inlabel() throws SJIOException
	{
		String lab = (String) readObjectNoCNFEWithDelCheck(); // Implicit dynamic type check. Move to an explicit check in socket state?

		if (RUNTIME_STATE)
		{
			state.inbranch(new SJLabel(lab));

			//SJRuntimeUtils.debugPrintln("Current state: " + state.currentState());
		}

		return lab;
	}

	public final boolean outsync(boolean bool) throws SJIOException
	{
		writeByte(SJ_BOOLEAN);
		writeBoolean(bool);

		if (RUNTIME_STATE)
		{
			state.outwhile(bool);

			//SJRuntimeUtils.debugPrintln("Current state: " + state.currentState());
		}

		return bool;
	}

	public final boolean insync() throws SJIOException
	{
		boolean bool = readBooleanWithDelegationCheck();

		if (RUNTIME_STATE)
		{
			state.inwhile(bool);

			//SJRuntimeUtils.debugPrintln("Current state: " + state.currentState());
		}

		return bool;
	}

	public final boolean recursionFalse(String lab) throws SJIOException
	{
		if (RUNTIME_STATE)
		{
			state.recursion(new SJLabel(lab));

			//SJRuntimeUtils.debugPrintln("Current state: " + state.currentState());
		}

		return false;
	}

	public final boolean recursionTrue(String lab) throws SJIOException
	{
		if (RUNTIME_STATE)
		{
			state.recurse(new SJLabel(lab));
		}

		return true;
	}

	private final void writeObject(Object obj) throws SJIOException
	{
		try
		{
			oos.writeObject(obj);
		}
		catch (IOException ioe)
		{
			throw new SJIOException(ioe);
		}
	}

	private final void writeBoolean(boolean v) throws SJIOException
	{
		try
		{
			oos.writeBoolean(v);
			oos.flush();
		}
		catch (IOException ioe)
		{
			throw new SJIOException(ioe);
		}
	}

	private final void writeInt(int v) throws SJIOException
	{
		try
		{
			oos.writeInt(v);
			oos.flush();
		}
		catch (IOException ioe)
		{
			throw new SJIOException(ioe);
		}
	}

	private final void writeByte(byte v) throws SJIOException // Rename writeFlag?
	{
		try
		{
			oos.writeByte(v); // Not flushed.
		}
		catch (IOException ioe)
		{
			throw new SJIOException(ioe);
		}
	}

	private final void writeByteAndFlush(byte v) throws SJIOException
	{
		try
		{
			oos.writeByte(v);
			oos.flush();
		}
		catch (IOException ioe)
		{
			throw new SJIOException(ioe);
		}
	}

	private final Object readObject() throws SJIOException, ClassNotFoundException
	{
		try
		{
			return ois.readObject();
		}
		catch (IOException ioe)
		{
			throw new SJIOException(ioe);
		}
	}

	private final boolean readBoolean() throws SJIOException
	{
		try
		{
			return ois.readBoolean();
		}
		catch (IOException ioe) // No need to propagate EOFException.
		{
			throw new SJIOException(ioe);
		}
	}

	private final int readInt() throws SJIOException
	{
		try
		{
			return ois.readInt();
		}
		catch (IOException ioe)
		{
			throw new SJIOException(ioe);
		}
	}

	private final byte readByte() throws SJIOException
	{
		try
		{
			return ois.readByte();
		}
		catch (IOException ioe)
		{
			throw new SJIOException(ioe);
		}
	}

	private final Object readObjectNoCNFE() throws SJIOException // Used when no need to check for delegation signal or read from lost message buffer (rename to reflect this?), and when failure due to an unavailabe class is equivalent to any other failure.
	{
		try
		{
			return readObject();
		}
		catch (ClassNotFoundException cnfe)
		{
			throw new SJIOException(cnfe);
		}
	}

	private final Object readObjectWithDelegationCheck() throws SJIOException, ClassNotFoundException // Also responsible for reading buffered lost messages (although delegation signals cannot be lost messages). Refactor to make this clear in the method name?
	{
		Object obj = null;

		if (lostMessages.size() == 0)
		{
			byte flag = readByte(); // Duals the "send protocol".

			if (flag == SJ_OBJECT || flag == SJ_START_DEL)
			{
				return readObject();
			}
			else if (flag == SJ_DEL_SIGNAL)
			{
				reconnectSingleDelegation((SJDelegationSignal) readObject());

				return readObjectWithDelegationCheck(); // Session could be re-delegated.
			}
			else //if (flag == SJ_FIN)
			{
				throw new SJIOException("Session prematurely closed by peer.");
			}
		}
		else
		{
			obj = readLostMessage();

			if (obj instanceof SJFIN) // Exception propagation. // Is this possible?
			{
				throw new SJIOException("Session prematurely closed by peer.");
			}

			return obj;
		}
	}

	private final Object readLostMessage() throws SJIOException, ClassNotFoundException
	{
		Object obj = lostMessages.remove(0);

		//SJRuntimeUtils.debugPrintln("Loaded lost message: " + obj);

		return obj;
	}

	private final boolean readBooleanWithDelegationCheck() throws SJIOException
	{
		if (lostMessages.size() == 0)
		{
			byte flag = readByte();

			if (flag == SJ_BOOLEAN)
			{
				return readBoolean();
			}
			else if (flag == SJ_DEL_SIGNAL)
			{
				reconnectSingleDelegation((SJDelegationSignal) readObjectNoCNFE());

				return readBooleanWithDelegationCheck();
			}
			else //if (flag == SJ_FIN)
			{
				throw new SJIOException("Session prematurely closed by peer.");
			}
		}
		else
		{
			try
			{
				return ((Boolean) readLostMessage()).booleanValue();
			}
			catch (ClassNotFoundException cnfe)
			{
				throw new SJIOException(cnfe);
			}
		}
	}

	private final int readIntWithDelegationCheck() throws SJIOException
	{
		if (lostMessages.size() == 0)
		{
			byte flag = readByte();

			if (flag == SJ_INT)
			{
				return readInt();
			}
			else if (flag == SJ_DEL_SIGNAL)
			{
				reconnectSingleDelegation((SJDelegationSignal) readObjectNoCNFE());

				return readIntWithDelegationCheck();
			}
			else //if (flag == SJ_FIN)
			{
				throw new SJIOException("Session prematurely closed by peer.");
			}
		}
		else
		{
			try
			{
				return ((Integer) readLostMessage()).intValue();
			}
			catch (ClassNotFoundException cnfe)
			{
				throw new SJIOException(cnfe);
			}
		}
	}

	private final byte readByteWithDelegationCheck() throws SJIOException
	{
		if (lostMessages.size() == 0)
		{
			byte flag = readByte();

			if (flag == SJ_BYTE)
			{
				return readByte();
			}
			else if (flag == SJ_DEL_SIGNAL)
			{
				reconnectSingleDelegation((SJDelegationSignal) readObjectNoCNFE());

				return readByteWithDelegationCheck();
			}
			else //if (flag == SJ_FIN)
			{
				throw new SJIOException("Session prematurely closed by peer.");
			}
		}
		else
		{
			try
			{
				return ((Byte) readLostMessage()).byteValue();
			}
			catch (ClassNotFoundException cnfe)
			{
				throw new SJIOException(cnfe);
			}
		}
	}

	private final Object readObjectNoCNFEWithDelCheck() throws SJIOException
	{
		try
		{
			return readObjectWithDelegationCheck();
		}
		catch (ClassNotFoundException cnfe)
		{
			throw new SJIOException(cnfe);
		}
	}

	public final SJType protocolType()
	{
		return protocolType;
	}

	public final SJType currentState()
	{
		return RUNTIME_STATE ? state.currentState() : null;
	}

	private final SJType expectedType()
	{
		return RUNTIME_STATE ? state.expectedType() : null;
	}

	protected final void init() throws SJIOException, SJIncompatibleSessionException // Can be moved to SJSocket? (Only if the required read and write methods are moved too.) // Maybe eager class loading and verification should also be done after a session is delegated. // Should verify compatible socket implementations.
	{
		SJType protocol = protocolType();

		try
		{
			boolean ourecd = eagerClassDownloadingEnabled();
			boolean ourecv = eagerClassVerificationEnabled();

			oos.writeBoolean(ourecd);
			oos.writeBoolean(ourecv);
			oos.flush(); // Any gains by manually flushing both booleans together?

			boolean theirecd = ois.readBoolean();
			boolean theirecv = ois.readBoolean();

			if (theirecd || theirecv)
			{
				//try // Collapsing try scopes will be faster?
				{
					writeObject(SJRuntimeUtils.getClassSVUIDs(protocol)); // Refactor and optimise for class verification. // Need to modify to include interfaces? // Could cache class SVUIDs map for future use.
				}
				/*catch (ClassNotFoundException cnfe)
				{
					throw new SJIOException(cnfe);
				}*/
			}

			if (ourecd || ourecv)
			{
				Map<String, Long> theirClassSVUIDs = null;

				theirClassSVUIDs = (Map<String, Long>) readObjectNoCNFE(); // No need to check for delegation.

				if (ourecd)
				{
					eagerClassDownload(protocol, theirClassSVUIDs);
				}

				if (ourecv)
				{
					checkClasses(theirClassSVUIDs);
				}
			}

			if (theirecv)
			{
				//verifyClasses(ourecv, theirecv, protocol, theirClassSVUIDs);

				Object ack = readObjectNoCNFE(); // Ack sent by checkClasses.

				if (!((String) ack).equals(CLASSES_VERIFIED))
				{
					throw new SJIncompatibleSessionException("IncompatibleClass: " + ack);
				}
			}

			if (CHECK_DUALITY)
			{
				checkDuality(protocol);
			}
		}
		catch (ClassNotFoundException cnfe)
		{
			throw new SJIOException(cnfe);
		}
		catch (IOException ioe)
		{
			throw new SJIOException(ioe);
		}
		catch (SJException x)
		{
			shutdown();

			if (x instanceof SJIncompatibleSessionException)
			{
				throw (SJIncompatibleSessionException) x;
			}
			else //if (x instanceof SJIOException)
			{
				throw (SJIOException) x;
			}
		}
	}

	private final void eagerClassDownload(SJType protocolType, Map<String, Long> theirMap) throws SJIOException
	{
		for (Entry<String, Long> e : theirMap.entrySet())
		{
			String className = e.getKey();
			Class c = null;

			try
			{
				c = Class.forName(className); // Any faster way to check this?
			}
			catch(ClassNotFoundException cnfe1)
			{
				try
				{
					c = getRemoteClassLoader().loadClass(className); // If we have just downloaded this class, then we implicitly do not need to verify it.
				}
				catch (ClassNotFoundException cnfe2)
				{
					throw new SJIOException(cnfe2);
				}

				//SJRuntimeUtils.debugPrintln("Eagerly loaded: " + c);
			}
		}
	}

	/*private final void verifyClasses(boolean ourecv, boolean theirecv, SJType protocolType, Map<String, Long> theirMap) throws SJIOException, SJIncompatibleSessionException
	{

	}*/

	private final void checkClasses(Map<String, Long> theirs) throws SJIOException,  SJIncompatibleSessionException
	{
		for (Entry<String, Long> e : theirs.entrySet())
		{
			String className = e.getKey();
			Class c = null;

			try
			{
				if (remoteClassLoadingEnabled()) // Codebase is set (not null), already checked.
				{
					c = Class.forName(className, true, getRemoteClassLoader());
				}
				else
				{
					c = Class.forName(className);
				}
			}
			catch(ClassNotFoundException cnfe) // Class not in codebase or remote class loading not enabled.
			{
				throw new SJIncompatibleSessionException("Could not verify class: " + className); // Actually can't get here: would have already failed trying to decode the peer's session type (all classes are needed to decode).
			}

			ObjectStreamClass osc = ObjectStreamClass.lookup(c);

			if (!(e.getValue().longValue() == osc.getSerialVersionUID()))
			{
				writeObject(className); // Report the bad class.

				throw new SJIncompatibleSessionException("Incompatible class: " + className);
			}

			//SJRuntimeUtils.debugPrintln("Eagerly checked class: " + className);
		}

		writeObject(CLASSES_VERIFIED);
	}

	private final void checkDuality(SJType ours) throws SJIOException, SJIncompatibleSessionException
	{
		writeObject(getChannel().getProtocol().encodedSJType());

		SJType theirs = decodeSJType((String) readObjectNoCNFE()); // Will throw NullPointerException if any classes are not available (and remote class loading oisabled or class not available from repository). // Possible to get premature SJ_FIN here? (Only if thread is interrupted?)

		if (!ours.treeDualtype(theirs))
		{
			throw new SJIncompatibleSessionException("Our session type \"" + ours + "\" incompatible with requested session type \"" + theirs + "\".");
		}

		//SJRuntimeUtils.debugPrintln("Session initiated: our type `" + ours + "', their type `" + theirs + "'");
	}
}
