/*
 * This file is part of SessionJ.
 *
 * Copyright (C) Raymond Hu, Imperial College London.
 *
 */

package sj.runtime.net;

import java.util.Stack;

import polyglot.types.SemanticException;
import polyglot.types.Type;

import sj.types.*;
import sj.types.typeobjects.sesstypes.*;
import sj.util.*;

import sj.runtime.*;
import sj.runtime.net.contexts.*;

// Analogous to SJSessionContext. But only tracks the session types for a single session, as each session has an individual underlying socket, with its own socket state object. // We currently record the type of the session completed so far. An alternative would be to track the concrete operation history: basically, loops unrolled. This might make the lost message resending algorithm simpler.
public final class SJSocketState
{
	private final SJRSocket rsocket; // HACK: only socket state knows when we're leaving a  recursion scope.
	private final SJTypeSystem sjts;
	private final SJType protocolType;

	private final Stack<SJSocketContext> contexts = new Stack<SJSocketContext>();
	//private final SJType lastResendState = null;

	public SJSocketState(SJTypeSystem sjts, SJType protocolType)
	{
		this.rsocket = null;
		this.sjts = sjts;
		this.protocolType = protocolType;//.treeClone();
	}

	public SJSocketState(SJRSocket rsocket, SJType protocolType)
	{
		this.rsocket = rsocket;
		this.sjts = rsocket.sjTypeSystem();
		this.protocolType = protocolType;//.treeClone();
	}

	public final SJType protocolType()
	{
		return protocolType;
	}

	public final SJSocketContext currentContext()
	{
		return contexts.peek();
	}

	private final SJType activeType()
	{
		return currentContext().activeType();
	}

	private final SJType implementedType()
	{
		return currentContext().implementedType();
	}

	public final boolean inIterationContext()
	{
		for (int i = contexts.size() - 1; i >= 1; i--)
		{
			if (contexts.get(i) instanceof SJIterationContext)
			{
				return true;
			}
		}

		return false;
	}

	public final SJType currentState() // i.e. the session type performed so far.
	{
		return currentState(0); // 0th is the top-level.
	}

	private final SJType currentState(int i)
	{
		SJSocketContext sjsc = contexts.get(i);
		SJType current = sjsc.implementedType();

		if (current == null)
		{
			if (i < contexts.size() - 1)
			{
				current = currentState(i + 1);
			}
		}
		else
		{
			current = current.treeClone();

			if (i < contexts.size() - 1)
			{
				current = current.append(currentState(i + 1));
			}
		}

		if (sjsc instanceof SJOutwhileContext) // Not possible during a delegation.
		{
			current = sjts.SJOutwhileType().body(current);
		}
		else if (sjsc instanceof SJInwhileContext)
		{
			current = sjts.SJInwhileType().body(current);
		}
		else if (sjsc instanceof SJOutbranchContext)
		{
			current = sjts.SJOutbranchType().branchCase(((SJOutbranchContext) sjsc).label(), current);
		}
		else if (sjsc instanceof SJInbranchContext)
		{
			current = sjts.SJInbranchType().branchCase(((SJInbranchContext) sjsc).label(), current);
		}
		else if (sjsc instanceof SJRecursionContext)
		{
			current = sjts.SJRecursionType(((SJRecursionContext) sjsc).label()).body(current);
		}

		return current;
	}

	public final SJType expectedType()
	{
		return activeType();
	}

	public final SJType accept() throws SJIOException
	{
		return begin();
	}

	public final SJType request() throws SJIOException
	{
		return begin();
	}

	private final SJType begin() throws SJIOException
	{
		SJBeginType sjbt = sjts.SJBeginType();
		SJType protocol = protocolType();

		if (!(protocol instanceof SJBeginType))
		{
			throw new SJIOException("Protool should start with `" + sjbt + "', not : " + protocol);
		}

		pushTopLevel(protocol);//.treeClone()); // `child' returns a cloned subtree.
		advanceContext(sjbt);

		return sjbt;//.nodeClone()); // (Node)clone is essentially equivalent to instantiation.
	}

	public final void open() // To receive delegated sessions.
	{
		pushTopLevel(protocolType());//.treeClone());
	}

	public final SJType send(Object obj) throws SJIOException // Also need a sendSession for type tracking.
	{
		SJSendType sjst = null;

		try
		{
			String m = fullClassName(obj);

			if (m.contains("[]"))
			{
				String n = m.substring(0, m.indexOf('['));

				Type messageType = null;

				if (n.equals("boolean")) // Factor out constants
				{
					messageType = sjts.Boolean();
				}
				else if (n.equals("int"))
				{
					messageType = sjts.Int();
				}
				else if (n.equals("byte"))
				{
					messageType = sjts.Byte();
				}
				else
				{
					messageType = sjts.typeForName(n);
				}

				int dims = (m.length() - m.indexOf('[')) / "[]".length();

				sjst = sjts.SJSendType(sjts.arrayOf(messageType, dims));
			}
			else
			{
				sjst = sjts.SJSendType(sjts.typeForName(m));
			}
		}
		catch (SemanticException se) // Not possible.
		{
			throw new SJIOException(se);
		}

		/*SJType active = activeType().nodeClone();

		if (!active.isSubtype(sjst)) // No point to check own actions.
		{
			throw new SJIOException("Expected `" + active + "' but implemented: " + sjst);
		}*/

		advanceContext(sjst);

		return sjst;//.nodeClone());
	}

	public final SJType sendBoolean(boolean v) throws SJIOException
	{
		SJSendType sjst = null;

		try
		{
			sjst = sjts.SJSendType(sjts.Boolean());
		}
		catch (SemanticException se) // Not possible.
		{
			throw new SJIOException(se);
		}

		/*SJType active = activeType().nodeClone();

		if (!active.isSubtype(sjst))
		{
			throw new SJIOException("Expected `" + active + "' but implemented: " + sjst);
		}*/

		advanceContext(sjst);

		return sjst;//.nodeClone());
	}

	public final SJType sendInt(int v) throws SJIOException
	{
		SJSendType sjst = null;

		try
		{
			sjst = sjts.SJSendType(sjts.Int());
		}
		catch (SemanticException se)
		{
			throw new SJIOException(se);
		}

		/*SJType active = activeType().nodeClone();

		if (!active.isSubtype(sjst))
		{
			throw new SJIOException("Expected `" + active + "' but implemented: " + sjst);
		}*/

		advanceContext(sjst);

		return sjst;//.nodeClone());
	}

	public final SJType sendByte(byte v) throws SJIOException
	{
		SJSendType sjst = null;

		try
		{
			sjst = sjts.SJSendType(sjts.Byte());
		}
		catch (SemanticException se)
		{
			throw new SJIOException(se);
		}

		/*SJType active = activeType().nodeClone();

		if (!active.isSubtype(sjst))
		{
			throw new SJIOException("Expected `" + active + "' but implemented: " + sjst);
		}*/

		advanceContext(sjst);

		return sjst;//.nodeClone());
	}

	public final SJType sendSession(SJType sjtype) throws SJIOException
	{
		SJSendType sjst = null;

		try
		{
			sjst = sjts.SJSendType(sjtype);
		}
		catch (SemanticException se)
		{
			throw new SJIOException(se);
		}

		/*SJType active = activeType().nodeClone();

		if (!active.isSubtype(sjst))
		{
			throw new SJIOException("Expected `" + active + "' but implemented: " + sjst);
		}*/

		advanceContext(sjst);

		return sjst;//.nodeClone());
	}

	public final SJType sendChannel(SJType sjtype) throws SJIOException
	{
		SJSendType sjst = null;

		try
		{
			sjst = sjts.SJSendType(sjtype);
		}
		catch (SemanticException se)
		{
			throw new SJIOException(se);
		}

		/*SJType active = activeType().nodeClone();

		if (!(sjtype instanceof SJBeginType) || !active.isSubtype(sjst))
		{
			throw new SJIOException("Expected `" + active + "' but implemented: " + sjst);
		}*/

		advanceContext(sjst);

		return sjst;//.nodeClone());
	}

	public final SJType receive(Object obj) throws SJIOException
	{
		SJReceiveType sjrt = null;

		try
		{
			String m = fullClassName(obj);

			if (m.contains("[]")) // Duplicated from send case.
			{
				String n = m.substring(0, m.indexOf('['));

				Type messageType = null;

				if (n.equals("boolean")) // Factor out constants
				{
					messageType = sjts.Boolean();
				}
				else if (n.equals("int"))
				{
					messageType = sjts.Int();
				}
				else if (n.equals("byte"))
				{
					messageType = sjts.Byte();
				}
				else
				{
					messageType = sjts.typeForName(n);
				}

				int dims = (m.length() - m.indexOf('[')) / "[]".length();

				sjrt = sjts.SJReceiveType(sjts.arrayOf(messageType, dims));
			}
			else
			{
				sjrt = sjts.SJReceiveType(sjts.typeForName(m));
			}
		}
		catch (SemanticException se) // Not possible.
		{
			throw new SJIOException(se);
		}

		/*SJType active = activeType();//.nodeClone();

		if (!sjrt.nodeSubtype(active))
		{
			throw new SJIOException("Expected `" + active + "' but received: " + sjrt);
		}*/

		advanceContext(sjrt);

		return sjrt;//.nodeClone());
	}

	public final SJType receiveBoolean(boolean v) throws SJIOException
	{
		SJReceiveType sjrt = null;

		try
		{
			sjrt = sjts.SJReceiveType(sjts.Boolean());
		}
		catch (SemanticException se) // Not possible.
		{
			throw new SJIOException(se);
		}

		/*SJType active = activeType();//.nodeClone();

		if (!sjrt.nodeSubtype(active))
		{
			throw new SJIOException("Expected `" + active + "' but received: " + sjrt);
		}*/

		advanceContext(sjrt);

		return sjrt;//.nodeClone());
	}

	public final SJType receiveInt(int v) throws SJIOException
	{
		SJReceiveType sjrt = null;

		try
		{
			sjrt = sjts.SJReceiveType(sjts.Int());
		}
		catch (SemanticException se) // Not possible.
		{
			throw new SJIOException(se);
		}

		/*SJType active = activeType();//.nodeClone();

		if (!sjrt.nodeSubtype(active))
		{
			throw new SJIOException("Expected `" + active + "' but received: " + sjrt);
		}*/

		advanceContext(sjrt);

		return sjrt;//.nodeClone());
	}

	public final SJType receiveByte(int v) throws SJIOException
	{
		SJReceiveType sjrt = null;

		try
		{
			sjrt = sjts.SJReceiveType(sjts.Byte());
		}
		catch (SemanticException se) // Not possible.
		{
			throw new SJIOException(se);
		}

		/*SJType active = activeType();//.nodeClone();

		if (!sjrt.nodeSubtype(active))
		{
			throw new SJIOException("Expected `" + active + "' but received: " + sjrt);
		}*/

		advanceContext(sjrt);

		return sjrt;//.nodeClone());
	}

	public final SJType receiveSession(SJType sjtype) throws SJIOException
	{
		SJReceiveType sjrt = null;

		try
		{
			sjrt = sjts.SJReceiveType(sjtype);
		}
		catch (SemanticException se) // Not possible.
		{
			throw new SJIOException(se);
		}

		/*SJType active = activeType();//.nodeClone();

		if (!sjrt.nodeSubtype(active))
		{
			throw new SJIOException("Expected `" + active + "' but received: " + sjrt);
		}*/

		advanceContext(sjrt);

		return sjrt;//.nodeClone());
	}

	public final SJType receiveChannel(SJType sjtype) throws SJIOException
	{
		SJReceiveType sjrt = null;

		try
		{
			sjrt = sjts.SJReceiveType(sjtype);
		}
		catch (SemanticException se) // Not possible.
		{
			throw new SJIOException(se);
		}

		/*SJType active = activeType();//.nodeClone();

		if (!(sjtype instanceof SJBeginType) || !sjrt.nodeSubtype(active))
		{
			throw new SJIOException("Expected `" + active + "' but received: " + sjrt);
		}*/

		advanceContext(sjrt);

		return sjrt;//.nodeClone());
	}

	public final SJType recurse(SJLabel lab) throws SJIOException // Recursion is "local" (so is checked by compiler), no dynamic check needed (no point to check own actions).
	{
		SJRecurseType sjrt = sjts.SJRecurseType(lab);

		advanceContext(sjrt);

		return sjrt;//.nodeClone());
	}

	public final void outbranch(SJLabel lab) throws SJIOException
	{
		SJType active = activeType();//.nodeClone(); // Clone only needed to get expected type.

		/*if (!(active instanceof SJInbranchType)) // No point to check own actions.
		{
			throw new SJIOException("Expected outbranch, but implemented: " + active);
		}*/

		pushOutbranch(lab, ((SJOutbranchType) active).branchCase(lab));
	}

	public final void inbranch(SJLabel lab) // Expected message was a String, already implicitly checked (casting) by socket implementation. Move check into here?
	{
		SJType active = activeType();//.nodeClone();

		/*if (!(active instanceof SJInbranchType))
		{
			throw new SJIOException("Expected inbranch, but implemented: " + active);
		}*/

		pushInbranch(lab, ((SJInbranchType) active).branchCase(lab));
	}

	public final void outwhile(boolean bool)  throws SJIOException // Could return performed type on false.
	{
		SJType active = activeType();//.nodeClone();

		if (active instanceof SJUnknownType) // In the innermost iteration, and it has an empty body...
		{
			popContext(); //...so we have to pop here, because there was no basic action to pop it.

			active = activeType();
		}
		/*else if (!(active instanceof SJOutwhileType)) // No point to check own actions.
		{
			throw new SJIOException("Expected outwhile, but implemented: " + active);
		}*/

		active = ((SJOutwhileType) active).body();

		if (bool)
		{
			pushOutwhile(active);
		}
		else
		{
			SJOutwhileType owt = sjts.SJOutwhileType().body(active); // Hacky? (The actual implemented type was lost when the context was popped by the last operation in the iteration, so have to use the expected type.)

			advanceContext(owt);
		}
	}

	public final void inwhile(boolean bool) throws SJIOException
	{
		SJType active = activeType();//.nodeClone();

		if (active instanceof SJUnknownType) // The innermost iteration has an empty body.
		{
			popContext();

			active = activeType();
		}
		/*else if (!(active instanceof SJInwhileType))
		{
			throw new SJIOException("Expected inwhile, but implemented: " + active);
		}*/

		active = ((SJInwhileType) active).body();

		if (bool)
		{
			pushInwhile(active);
		}
		else
		{
			SJInwhileType iwt = sjts.SJInwhileType().body(active);

			advanceContext(iwt);
		}
	}

	public final void recursion(SJLabel lab) throws SJIOException // Recursion is "local" (so is checked by compiler), no dynamic check needed (no point to check own actions).
	{
		pushRecursion(lab, ((SJRecursionType) activeType()).body());
	}

	public final void close()
	{
		reset();
	}

	public final void reset()
	{
		contexts.removeAllElements();
	}

	private final void advanceContext(SJType sjtype)
	{
		SJType implemented = sjtype;

		//for (boolean redo = true; redo; )
		while (true)
		{
			SJSocketContext sjsc = currentContext();
			SJType completed = appendToImplemented(implemented);//.nodeClone()); // Implemented is the single operation just performed.
			SJType next = sjsc.activeType().child(); // Maybe end types should be (implicitly) reintroduced.

			sjsc.activeType(next);

			//SJRuntimeUtils.debugPrintln("Socket state: " + activeType() + ", " + implementedType());

			if (next == null && !(sjsc instanceof SJTopLevelContext))
			{
				if (sjsc instanceof SJIterationContext)
				{
					if (sjsc instanceof SJOutwhileContext || sjsc instanceof SJInwhileContext)
					{
						popContext(); // sjsc popped. // The iteration body type actually implemented is lost.

						break;//redo = false;
					}
					else if (sjsc instanceof SJRecursionContext)
					{
						if (sjtype instanceof SJRecurseType) // Enacting the recursion.
						{
							SJLabel lab = ((SJRecurseType) sjtype).label();

							while (!((SJRecursionContext) currentContext()).label().equals(lab))
							{
								popContext();
							}

							popContext(); // The recursion body type actually implemented is lost.

							break;//redo = false;
						}
						else // This recursion (including "nested mutual" scopes) has finished.
						{
							popContext(); // sjsc popped.

							SJLabel lab = ((SJRecursionContext) sjsc).label();

							implemented = sjts.SJRecursionType(lab).body(completed); // Only the branch type that quits the iteration is recorded.

							if (rsocket != null) // HACK: leaving recursion context.
							{
								if (!inIterationContext())
								{
									rsocket.pushSentMessage(SJRSocket.CACHE_MARKER);
								}
							}
						}
					}
				}
				else if (sjsc instanceof SJBranchContext)
				{
					popContext(); // sjsc popped.

					SJLabel lab = ((SJBranchContext) sjsc).label();

					if (sjsc instanceof SJOutbranchContext)
					{
						//advanceContext(sjts.SJOutbranchType().branchCase(lab, completed));
						implemented = sjts.SJOutbranchType().branchCase(lab, completed);
					}
					else //if (sjsc instanceof SJInbranchContext)
					{
						//advanceContext(sjts.SJInbranchType().branchCase(lab, completed));
						implemented = sjts.SJInbranchType().branchCase(lab, completed);
					}
				}
			}
			else
			{
				break;//redo = false;
			}
		}
	}

	private final SJType appendToImplemented(SJType sjtype)
	{
		SJType implemented = implementedType();

		if (implemented != null)
		{
			sjtype = implemented.append(sjtype);
		}

		currentContext().implementedType(sjtype);

		return sjtype;
	}

	public final void pushTopLevel(SJType sjtype)
	{
		pushContext(new SJTopLevelContext(sjtype));
	}

	public final void pushOutbranch(SJLabel lab, SJType sjtype)
	{
		if (sjtype == null) // Possible mismatch between program execution and session execution if an exception is raised within the (empty) branch. Does it matter? // So all possible branch cases are recorded, not the ones that were actually taken.
		{
			advanceContext(sjts.SJOutbranchType().branchCase(lab, null));
		}
		else
		{
			pushContext(new SJOutbranchContext(lab, sjtype));
		}
	}

	public final void pushInbranch(SJLabel lab, SJType sjtype)
	{
		if (sjtype == null)
		{
			advanceContext(sjts.SJInbranchType().branchCase(lab, null));
		}
		else
		{
			pushContext(new SJInbranchContext(lab, sjtype));
		}
	}

	public final void pushOutwhile(SJType sjtype)
	{
		if (sjtype == null)
		{
			pushContext(new SJOutwhileContext(sjts.SJUnknownType())); // Hacky? To represent empty iteration.
		}
		else
		{
			pushContext(new SJOutwhileContext(sjtype));
		}
	}

	public final void pushInwhile(SJType sjtype)
	{
		if (sjtype == null)
		{
			pushContext(new SJInwhileContext(sjts.SJUnknownType()));
		}
		else
		{
			pushContext(new SJInwhileContext(sjtype));
		}
	}

	public final void pushRecursion(SJLabel lab, SJType sjtype)
	{
		if (sjtype == null)
		{
			advanceContext(sjts.SJRecursionType(lab).body(null)); // Like branch: an empty recursion can't recurse because no recurse type in body.
		}
		else
		{
			pushContext(new SJRecursionContext(lab, sjtype));
		}
	}

	private final void pushContext(SJSocketContext sjsc)
	{
		contexts.push(sjsc);

		//SJRuntimeUtils.debugPrintln("Pushed " + sjsc + ": " + activeType() + ", " + implementedType());
	}

	private final void popContext()
	{
		SJSocketContext sjsc = contexts.pop();

		//SJRuntimeUtils.debugPrintln("Popped " + sjsc + " to: " + activeType() + ", " + implementedType());
	}

	private final String fullClassName(Object obj) // Move to SJRuntimeUtils?
	{
		String m = obj.getClass().toString().substring("class ".length());

		if (m.startsWith("["))
		{
			int depth = m.lastIndexOf('[') + 1;

			m = m.substring(m.lastIndexOf('[') + 1);

			if (m.equals("Z")) // See JLS:Arrays.
			{
				m = "boolean";
			}
			else if (m.equals("I"))
			{
				m = "int";
			}
			else if (m.equals("B"))
			{
				m = "byte";
			}
			else if (m.startsWith("L")) // Non-primitive type array.
			{
				m = m.substring("L".length());
				m = m.substring(0, m.length() - ";".length());
			}
			else
			{
				System.out.println("Unrecognised array class signature: " + m);
			}

			for ( ; depth > 0; depth--)
			{
				m = m + "[]";
			}
		}
		else
		{

		}

		return m;
	}
}
