/*
 * This file is part of SessionJ.
 *
 * Copyright (C) Raymond Hu, Imperial College London.
 *
 */

package sj.runtime.util;

import java.io.*;
import java.net.*;
import java.util.*;

import polyglot.types.*;

import sj.types.typeobjects.sesstypes.*;
import sj.util.SJLabel;

import sj.runtime.*;

public final class SJRuntimeUtils
{
	private static final boolean debugMode = true;

	private static final int LOWER_PORT_LIMIT = 1024; // Move to SJConstants?
	private static final int PORT_SEED = 10000;

	private static Random random = new Random();

	private SJRuntimeUtils() { }

	public static final void debugPrint(Object obj)
	{
		if (debugMode)
		{
			System.out.print(obj);
		}
	}

	public static final void debugPrintln(Object obj)
	{
		if (debugMode)
		{
			System.out.println(obj);
		}
	}

	/*public static SJTypeSystem SJTypeSystem() throws SemanticException
	{
		ExtensionInfo extInfo = new ExtensionInfo();
		SJTypeSystem sjts = (SJTypeSystem) extInfo.typeSystem();
		TopLevelResolver tlr = new SJClassResolver(sjts, extInfo);

		sjts.initialize(tlr, extInfo);

		return sjts;
	}*/

	public static Map<String, Long> getClassSVUIDs(SJType sjtype) throws SJIOException, ClassNotFoundException
	{
		HashMap<String, Long> ours = new HashMap<String, Long>();

		while (sjtype != null)
		{
			if (sjtype instanceof SJOutbranchType)
			{
				SJOutbranchType obt = (SJOutbranchType) sjtype;

				for (SJLabel lab : obt.labelSet())
				{
					ours.putAll(getClassSVUIDs(obt.branchCase(lab)));
				}
			}
			else if (sjtype instanceof SJInbranchType)
			{
				SJInbranchType ibt = (SJInbranchType) sjtype;

				for (SJLabel lab : ibt.labelSet())
				{
					ours.putAll(getClassSVUIDs(ibt.branchCase(lab)));
				}
			}
			else if (sjtype instanceof SJOutwhileType)
			{
				ours.putAll(getClassSVUIDs(((SJOutwhileType) sjtype).body()));
			}
			else if (sjtype instanceof SJInwhileType)
			{
				ours.putAll(getClassSVUIDs(((SJInwhileType) sjtype).body()));
			}
			else if (sjtype instanceof SJRecursionType)
			{
				ours.putAll(getClassSVUIDs(((SJRecursionType) sjtype).body()));
			}
			else if (sjtype instanceof SJSendType)
			{
				ours.putAll(getMessageSVUID(((SJSendType) sjtype).messageType()));
			}
			else if (sjtype instanceof SJReceiveType)
			{
				ours.putAll(getMessageSVUID(((SJReceiveType) sjtype).messageType()));
			}
			else if (sjtype instanceof SJRecurseType)
			{
				// Nothing to do.
			}

			sjtype = sjtype.child();
		}

		return ours;
	}

	private static Map<String, Long> getMessageSVUID(Type messageType) throws SJIOException, ClassNotFoundException // SJCompilerUtils has a simpler version of this routine.
	{
		HashMap<String, Long> ours = new HashMap<String, Long>();

		if (messageType instanceof SJType) // Should come before ordinary class cases? (But SJType shouldn't be a class type).
		{
			ours.putAll(getClassSVUIDs((SJType) messageType));
		}
		else if (messageType.isPrimitive())
		{
			// No SVUID needed for primitive types.
		}
		else if (messageType.isClass()) // Duplicated from above.
		{
			String className = messageType.toClass().fullName();
			Class<?> c = Class.forName(className);

			if (c.isInterface())
			{
				// Interfaces don't have SVUIDs (so what should we do here?). // This encourages use of abstract classes rather than interfaces for message types?
			}
			else
			{
				ObjectStreamClass osc = ObjectStreamClass.lookup(c);

				if (osc == null)
				{
					throw new SJIOException("Class not serializable: " + c);
				}

				ours.put(className, osc.getSerialVersionUID()); // Not possible to find a different SVUID for the same (name) class here? // SVUIDs could be recorded in the session type objects. // Put currently problems working with these values at compilation time if the class binary is not available a priori (SVUID value not built yet?).
			}
		}
		else if (messageType.isArray())
		{
			throw new SJIOException("Array types not done yet: " + messageType);
		}

		return ours;
	}

	public static int findFreeTCPPort()
	{
		int port = LOWER_PORT_LIMIT + random.nextInt(PORT_SEED); // Make the Random a static? Or take the port seed as a parameter?

		for (boolean redo = true; redo; )
		{
			ServerSocket ss = null;

			try
			{
				ss = new ServerSocket(port);

				redo = false;
			}
			catch (IOException ioe)
			{
				port++; // No upper port limit specified (assumes we will eventually find a valid free port).
			}
			finally
			{
				try
				{
					closeServerSocket(ss);
				}
				catch (IOException ioe)
				{
					// Any danger?
				}
			}
		}

		return port;
	}

	// Move to SJRemoteClassHandler?
	public static void closeStream(InputStream is) throws IOException
	{
		if (is != null)
		{
			is.close();
		}
	}

	public static void closeStream(OutputStream os) throws IOException
	{
		if (os != null)
		{
			os.flush(); // Is this useful?
			os.close();
		}
	}

	public static void closeSocket(Socket s) throws IOException
	{
		if (s != null)
		{
			s.close();
		}
	}

	public static void closeServerSocket(ServerSocket ss) throws IOException
	{
		if (ss != null)
		{
			ss.close();
		}
	}
}
